#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import functools
from datetime import date
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnHeader, ColumnID, Filter, FilterGroup, OperatorType, StepType
from mitosheet.types import (
    FC_BOOLEAN_IS_FALSE, FC_BOOLEAN_IS_TRUE, FC_DATETIME_EXACTLY,
    FC_DATETIME_GREATER, FC_DATETIME_GREATER_THAN_OR_EQUAL, FC_DATETIME_LESS,
    FC_DATETIME_LESS_THAN_OR_EQUAL, FC_DATETIME_NOT_EXACTLY, FC_EMPTY,
    FC_LEAST_FREQUENT, FC_MOST_FREQUENT, FC_NOT_EMPTY, FC_NUMBER_EXACTLY,
    FC_NUMBER_GREATER, FC_NUMBER_GREATER_THAN_OR_EQUAL, FC_NUMBER_HIGHEST,
    FC_NUMBER_LESS, FC_NUMBER_LESS_THAN_OR_EQUAL, FC_NUMBER_LOWEST,
    FC_NUMBER_NOT_EXACTLY, FC_STRING_CONTAINS, FC_STRING_DOES_NOT_CONTAIN,
    FC_STRING_ENDS_WITH, FC_STRING_EXACTLY, FC_STRING_NOT_EXACTLY,
    FC_STRING_STARTS_WITH, FC_STRING_CONTAINS_CASE_INSENSITIVE)


# If there are multiple conditions, we combine them together, with the
# given operator in the middle
OPERATOR_SIGNS = {"Or": "|", "And": "&"}

# Filter conditions that cannot be applied to the first 1500 rows of the dataframe 
# should be put here. They require different handling in conditonal formats, for example
FILTER_CONDITIONS_THAT_REQUIRE_FULL_DATAFRAME = [
    FC_LEAST_FREQUENT,
    FC_MOST_FREQUENT,
    FC_NUMBER_LOWEST,
    FC_NUMBER_HIGHEST,
]

class FilterStepPerformer(StepPerformer):
    """
    Allows you to filter a column based on some conditions and some values.
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return "filter_column"

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:
        """
        Saturates the filter event with a `has_non_empty_filter` - which is useful
        for for logging
        """
        has_non_empty_filter = False
        for filter_or_group in params["filters"]:
            if "filters" in filter_or_group:
                # If it's a group
                if len(filter_or_group["filters"]) > 0:
                    has_non_empty_filter = True
            else:
                # If it's a single filter
                has_non_empty_filter = True

        params["has_non_empty_filter"] = has_non_empty_filter
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        post_state, execution_data = cls.execute_through_transpile(prev_state, params)

        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        operator: OperatorType = get_param(params, 'operator')
        filters: Any = get_param(params, 'filters')

        # Keep track of which columns are filtered
        post_state.column_filters[sheet_index][column_id]["operator"] = operator
        post_state.column_filters[sheet_index][column_id]["filters"] = filters

        return post_state, execution_data

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        from mitosheet.code_chunks.step_performers.filter_code_chunk import \
            FilterCodeChunk
        return [
            FilterCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                [{
                    'column_id': get_param(params, 'column_id'),
                    'filter': {
                        'operator': get_param(params, 'operator'),
                        'filters': get_param(params, 'filters')
                    }
                }]
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


def get_applied_filter(
    df: pd.DataFrame, column_header: ColumnHeader, filter_: Filter
) -> pd.Series:
    """
    Given a filter triple, returns the filter indexes for that
    actual dataframe
    """
    condition = filter_["condition"]
    value = filter_["value"]

    # First, check shared filter conditions
    if condition == FC_EMPTY:
        return df[column_header].isna()
    elif condition == FC_NOT_EMPTY:
        return df[column_header].notnull()
    elif condition == FC_LEAST_FREQUENT:
        value_int: int = value # type: ignore
        return df[column_header].isin(df[column_header].value_counts().index.tolist()[-value_int:])
    elif condition == FC_MOST_FREQUENT:
        value_int: int = value # type: ignore
        return df[column_header].isin(df[column_header].value_counts().index.tolist()[:value_int])


    # Then bool
    if condition == FC_BOOLEAN_IS_TRUE:
        return df[column_header] == True
    elif condition == FC_BOOLEAN_IS_FALSE:
        return df[column_header] == False

    # Then string
    if condition == FC_STRING_CONTAINS:
        return df[column_header].str.contains(value, na=False, regex=False)
    elif condition == FC_STRING_DOES_NOT_CONTAIN:
        return ~df[column_header].str.contains(value, na=False, regex=False)
    elif condition == FC_STRING_STARTS_WITH:
        return df[column_header].str.startswith(value, na=False)
    elif condition == FC_STRING_ENDS_WITH:
        return df[column_header].str.endswith(value, na=False)
    elif condition == FC_STRING_EXACTLY:
        return df[column_header] == value
    elif condition == FC_STRING_NOT_EXACTLY:
        return df[column_header] != value
    elif condition == FC_STRING_CONTAINS_CASE_INSENSITIVE:
        return df[column_header].str.contains(value, na=False, regex=False, case=False)

    # Then number
    if condition == FC_NUMBER_EXACTLY:
        return df[column_header] == value
    elif condition == FC_NUMBER_NOT_EXACTLY:
        return df[column_header] != value
    elif condition == FC_NUMBER_GREATER:
        return df[column_header] > value
    elif condition == FC_NUMBER_GREATER_THAN_OR_EQUAL:
        return df[column_header] >= value
    elif condition == FC_NUMBER_LESS:
        return df[column_header] < value
    elif condition == FC_NUMBER_LESS_THAN_OR_EQUAL:
        return df[column_header] <= value
    elif condition == FC_NUMBER_LOWEST:
        return df[column_header].isin(df[column_header].nsmallest(value, keep='all'))
    elif condition == FC_NUMBER_HIGHEST:
        return df[column_header].isin(df[column_header].nlargest(value, keep='all'))

    # Check that we were given something that can be understood as a date
    try:
        timestamp = pd.to_datetime(value)
    except:
        # If we hit an error, because we restrict the input datetime,
        # this is probably occuring because the user has only partially input the date,
        # and so in this case, we just default it to the minimum possible timestamp for now!
        timestamp = date.min

    if condition == FC_DATETIME_EXACTLY:
        return df[column_header] == timestamp
    elif condition == FC_DATETIME_NOT_EXACTLY:
        return df[column_header] != timestamp
    elif condition == FC_DATETIME_GREATER:
        return df[column_header] > timestamp
    elif condition == FC_DATETIME_GREATER_THAN_OR_EQUAL:
        return df[column_header] >= timestamp
    elif condition == FC_DATETIME_LESS:
        return df[column_header] < timestamp
    elif condition == FC_DATETIME_LESS_THAN_OR_EQUAL:
        return df[column_header] <= timestamp

    raise Exception(f"Invalid type passed in filter {filter_}")


def combine_filters(operator: OperatorType, filters: List[pd.Series]) -> pd.Series:
    def filter_reducer(filter_one: pd.Series, filter_two: pd.Series) -> pd.Series:
        # Helper for combining filters based on the operations
        if operator == "Or":
            return (filter_one) | (filter_two)
        elif operator == "And":
            return (filter_one) & (filter_two)
        else:
            raise Exception(f"Operator {operator} is unsupported")

    # Combine all the filters into a single filter
    return functools.reduce(filter_reducer, filters)

def get_full_applied_filter(
    df: pd.DataFrame,
    column_header: ColumnHeader,
    operator: OperatorType,
    filters: List[Union[Filter, FilterGroup]],
) -> Tuple[pd.Series, float]:
    applied_filters = []
    pandas_start_time = perf_counter()

    for filter_or_group in filters:
        if "filters" not in filter_or_group:
            filter_: Filter = filter_or_group #type: ignore
            applied_filters.append(get_applied_filter(df, column_header, filter_))
        else:
            filter_group: FilterGroup = filter_or_group #type: ignore
            full_group_filter, _ = get_full_applied_filter(df, column_header, filter_group['operator'], filter_group["filters"])
            applied_filters.append(full_group_filter)

    if len(applied_filters) > 0:
        full_applied_filter = combine_filters(operator, applied_filters)
    else:
        full_applied_filter = pd.Series(data=True, index=df.index, dtype='bool')
    
    pandas_processing_time = perf_counter() - pandas_start_time

    return (full_applied_filter, pandas_processing_time)

def check_filters_contain_condition_that_needs_full_df(filters: List[Union[Filter, FilterGroup]]) -> bool:
    """
    Returns true if any filter condition is a FILTER_CONDITIONS_THAT_REQUIRE_FULL_DATAFRAME
    """

    for filter_or_group in filters:

        if 'filters' not in filter_or_group:
            filter_: Filter = filter_or_group #type: ignore
            if filter_['condition'] in FILTER_CONDITIONS_THAT_REQUIRE_FULL_DATAFRAME:
                return True
        else:
            filter_group: FilterGroup = filter_or_group #type: ignore
            if check_filters_contain_condition_that_needs_full_df(filter_group["filters"]):
                return True
        
    return False



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def update_filter_column_1_to_filter_column_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Upgrades from a filter column version 1 step to version 2. This was
    part of the refactor that unified type handling across the codebase,
    and got rid of an old distinction between a mito column type and a filter
    type. Now, we just have mito types!

    Old format of the step: {
        "step_version": 1, 
        "step_type": "filter_column", 
        "sheet_index": 0, 
        "column_header": "A", 
        "filters": [
            {"type": "number", "condition": "greater", "value": 1}, 
            {"filters": [
                {"type": "number", "condition": "greater", "value": 2}, 
                {"type": "number", "condition": "greater", "value": 3}], 
                "operator": "And"}
            ], 
        "operator": "And"}
    }

    New version of the step: {
        "step_version": 2, 
        "step_type": "filter_column", 
        "sheet_index": 0, 
        "column_header": "A", 
        "filters": [
            {"type": "number", "condition": "greater", "value": 1}, 
            {"filters": [
                {"type": "number", "condition": "greater", "value": 2}, 
                {"type": "number", "condition": "greater", "value": 3}], 
                "operator": "And"}
            ], 
        "operator": "And"}
    }

    Anywhere it says "type", we change from:
    - number -> number_series
    - string -> string_series
    - datetime -> datetime_series
    """
    type_mapping = {
        'string': 'string_series',
        'number': 'number_series',
        'datetime': 'datetime_series',
    }

    new_filters = []

    for filter_or_group in step['filters']:
        # We check if this is a filter or a group, and 
        # case appropriately!
        if 'filters' in filter_or_group:
            # Filter group case
            group = filter_or_group
            new_filters_in_group = []
            for filter_ in group['filters']:
                filter_['type'] = type_mapping[filter_['type']]
                new_filters_in_group.append(filter_)
            group['filters'] = new_filters_in_group

            new_filters.append(filter_or_group)
        else:
            # Just a regular filter, so rename for clarity
            filter_ = filter_or_group
            filter_['type'] = type_mapping[filter_['type']]
            new_filters.append(filter_)

    return [{
        "step_version": 2, 
        "step_type": "filter_column", 
        "sheet_index": step['sheet_index'], 
        "column_header": step['column_header'], 
        "filters": new_filters, 
        "operator": step['operator']
    }] + later_steps


def upgrade_filter_column_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        "step_version": 2, 
        "step_type": "filter_column", 
        "params": {
            "sheet_index": 0, 
            "column_header": "A", 
            "filters": [
                {"type": "number", "condition": "greater", "value": 1}, 
                {"filters": [
                    {"type": "number", "condition": "greater", "value": 2}, 
                    {"type": "number", "condition": "greater", "value": 3}], 
                    "operator": "And"}
                ], 
            "operator": "And"}
        }
    }

    NEW: {
        "step_version": 3, 
        "step_type": "filter_column", 
        "params": {
            "sheet_index": 0, 
            "column_id": _id_, 
            "filters": [
                {"type": "number", "condition": "greater", "value": 1}, 
                {"filters": [
                    {"type": "number", "condition": "greater", "value": 2}, 
                    {"type": "number", "condition": "greater", "value": 3}], 
                    "operator": "And"}
                ], 
            "operator": "And"}
        }
    }
    # NOTE: The above filter group look copied and pasted wrong, but not gonna edit
    # it given that these upgraders should be immutable.
    """
    params = step['params']
    params = replace_headers_with_id(params, 'column_header', 'column_id')

    return [{
        "step_version": 3, 
        "step_type": "filter_column", 
        "params": params
    }] + later_steps


def upgrade_filter_column_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Removes all notion of a type from a filter, as it is totally not necessary
    as we removed the notion of a Mito Type

    OLD: {
        "step_version": 3, 
        "step_type": "filter_column", 
        "params": {
            "sheet_index": 0, 
            "column_id": _id_, 
            "filters": [
                {"type": "number", "condition": "greater", "value": 1}, 
                {"filters": [
                    {"type": "number", "condition": "greater", "value": 2}, 
                    {"type": "number", "condition": "greater", "value": 3}], 
                    "operator": "And"}
                ], 
            "operator": "And"
        }
    }
    NEW: {
        "step_version": 4, 
        "step_type": "filter_column", 
        "params": {
            "sheet_index": 0, 
            "column_id": _id_, 
            "filters": [
                {"condition": "greater", "value": 1}, 
                {
                    "filters": [
                        {"condition": "greater", "value": 2}, 
                        {"condition": "greater", "value": 3}], 
                    ], 
                    "operator": "And"
                }
            "operator": "And"}
        }
    }
    """

    params = step['params']
    new_filters = []

    for filter_or_group in params['filters']:
        # We check if this is a filter or a group, and case appropriately!
        if 'filters' in filter_or_group:
            # Filter group case
            group = filter_or_group
            new_filters_in_group = []
            for filter_ in group['filters']:
                del filter_['type']
                new_filters_in_group.append(filter_)
            group['filters'] = new_filters_in_group

            new_filters.append(filter_or_group)
        else:
            # Just a regular filter, so rename for clarity
            filter_ = filter_or_group
            del filter_['type']
            new_filters.append(filter_)

    params['filters'] = new_filters

    return [{
        "step_version": 4, 
        "step_type": "filter_column", 
        "params": params
    }] + later_steps



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that are useful for control flow. 

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""

from typing import Optional

import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.sheet_functions.utils import (
    get_final_result_series_or_primitive, get_series_from_primitive_or_series)
from mitosheet.public.v3.types.decorators import (
    cast_values_in_all_args_to_type, cast_values_in_arg_to_type)
from mitosheet.public.v3.types.sheet_function_types import (
    AnyPrimitiveOrSeriesInputType, BoolFunctionReturnType, BoolInputType,
    BoolRestrictedInputType, IfsInputType)


@cast_values_in_all_args_to_type('bool')
@handle_sheet_function_errors
def AND(*argv: Optional[BoolInputType]) -> BoolFunctionReturnType:
    """
    {
        "function": "AND",
        "description": "Returns True if all of the provided arguments are True, and False if any of the provided arguments are False.",
        "category": "LOGIC",
        "search_terms": ["and", "&", "if", "conditional"],
        "examples": [
            "AND(True, False)",
            "AND(Nums > 100, Nums < 200)",
            "AND(Pay > 10, Pay < 20, Status == 'active')"
        ],
        "syntax": "AND(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    # If we don't find any arguements, we default to True. Excel in practice
    # defaults to a value error if there are no args
    return get_final_result_series_or_primitive(
        True,
        argv,
        lambda df: df.all().all(),
        lambda previous_value, new_value: previous_value and new_value,
        lambda previous_series, new_series: previous_series & new_series
    )

@cast_values_in_arg_to_type('series', 'bool')
@handle_sheet_function_errors
def BOOL(series: BoolRestrictedInputType) -> BoolFunctionReturnType:
    """
    {
        "function": "BOOL",
        "description": "Converts the passed arguments to boolean values, either True or False. For numberic values, 0 converts to False while all other values convert to True.",
        "search_terms": ["bool", "boolean", "true", "false", "dtype", "convert"],
        "category": "LOGIC",
        "examples": [
            "BOOL(Amount_Payed)",
            "AND(BOOL(Amount_Payed), Is_Paying)"
        ],
        "syntax": "BOOL(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "An series to convert to boolean values, either True or False."
            }
        ]
    }
    """
    if isinstance(series, bool):
        return series

    return series.fillna(False)

@cast_values_in_arg_to_type('condition', 'bool')
@handle_sheet_function_errors
def IF(condition: pd.Series, true_series: AnyPrimitiveOrSeriesInputType, false_series: AnyPrimitiveOrSeriesInputType) -> pd.Series:
    """
    {
        "function": "IF",
        "description": "Returns one value if the condition is True. Returns the other value if the conditon is False.",
        "search_terms": ["if", "conditional", "and", "or"],
        "category": "LOGIC",
        "examples": [
            "IF(Status == 'success', 1, 0)",
            "IF(Nums > 100, 100, Nums)",
            "IF(AND(Grade >= .6, Status == 'active'), 'pass', 'fail')"
        ],
        "syntax": "IF(boolean_condition, value_if_true, value_if_false)",
        "syntax_elements": [{
                "element": "boolean_condition",
                "description": "An expression or series that returns True or False values. Valid conditions for comparison include ==, !=, >, <, >=, <=."
            },
            {
                "element": "value_if_true",
                "description": "The value the function returns if condition is True."
            },
            {
                "element": "value_if_false",
                "description": "The value the function returns if condition is False."
            }
        ]
    }
    """

    true_series = get_series_from_primitive_or_series(true_series, condition.index)
    false_series = get_series_from_primitive_or_series(false_series, condition.index)

    return pd.Series(
        data=[true_series.loc[i] if c else false_series.loc[i] for i, c in condition.items()],
        index=condition.index
    )


@handle_sheet_function_errors
def IFS(*argv: Optional[IfsInputType]) -> pd.Series:
    """
    {
        "function": "IFS",
        "description": "Returns the value of the first condition that is true. If no conditions are true, returns None.",
        "search_terms": ["ifs", "if", "conditional", "and", "or"],
        "category": "LOGIC",
        "examples": [
            "IFS(height > 100, 'tall', height > 50, 'medium', height > 0, 'short')"
        ],
        "syntax": "IFS(boolean_condition_1, value_if_true, [boolean_condition_2, value_if_true, ...])",
        "syntax_elements": [{
                "element": "boolean_condition",
                "description": "An expression or series that returns True or False values. Valid conditions for comparison include ==, !=, >, <, >=, <=."
            },
            {
                "element": "value_if_true, ... [OPTIONAL]",
                "description": "The value the function returns if condition is True, followed by alternating boolean conditions and values."
            }
        ]
    }
    """
    if len(argv) % 2 != 0:
        raise MitoError(
            'invalid_args_error',
            'IFS',
            'IFS requires an even number of arguments.',
            error_modal=False
        )
    base_index = next(iter(s.index for s in argv if isinstance(s, pd.Series)), None)
    # If index is None, we're dealing with all constants
    if base_index is None:
        for index, arg in enumerate(argv):
            if index % 2 == 0:
                if arg == True:
                    return argv[index+1]
        return None

    else:
        # Otherwise, we have at least one series -- so we can go through and turn all of the constants into series.
        argv_series = tuple([get_series_from_primitive_or_series(arg, base_index) for arg in argv])
        results = pd.Series(index=base_index)

        for index, condition in enumerate(argv_series):
            if index % 2 == 0:
                if condition.dtype != bool:
                    raise MitoError(
                        'invalid_args_error',
                        'IFS',
                        f"IFS requires all even indexed arguments to be boolean.",
                        error_modal=False
                    )
                
                # If it is, use the "true_series" to fill the value in the result series
                true_series = argv_series[index+1]
                new_series = true_series[condition]
                results = results.combine_first(new_series)
                
    return results

@cast_values_in_all_args_to_type('bool')
@handle_sheet_function_errors
def OR(*argv: Optional[BoolInputType]) -> BoolFunctionReturnType:
    """
    {
        "function": "OR",
        "description": "Returns True if any of the provided arguments are True, and False if all of the provided arguments are False.",
        "search_terms": ["or", "if", "conditional"],
        "category": "LOGIC",
        "examples": [
            "OR(True, False)",
            "OR(Status == 'success', Status == 'pass', Status == 'passed')"
        ],
        "syntax": "OR(boolean_condition1, [boolean_condition2, ...])",
        "syntax_elements": [{
                "element": "boolean_condition1",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            },
            {
                "element": "boolean_condition2 ... [OPTIONAL]",
                "description": "An expression or series that returns True or False values. See IF documentation for a list of conditons."
            }
        ]
    }
    """

    # If we don't find any arguements, we default to True. Excel in practice
    # defaults to a value error if there are no args
    return get_final_result_series_or_primitive(
        False,
        argv,
        lambda df: df.any().any(),
        lambda previous_value, new_value: previous_value or new_value,
        lambda previous_series, new_series: previous_series | new_series
    )


# TODO: we should see if we can list these automatically!
CONTROL_FUNCTIONS = {
    'AND': AND,
    'BOOL': BOOL,
    'IF': IF,
    'IFS': IFS,
    'OR': OR,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.drop_duplicates_code_chunk import DropDuplicatesCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


class DropDuplicatesStepPerformer(StepPerformer):
    """
    Allows you to drop duplicates from a dataframe
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'drop_duplicates'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index = get_param(params, 'sheet_index')

        post_state, execution_data = cls.execute_through_transpile(prev_state, params)

        num_rows_dropped = len(prev_state.dfs[sheet_index].index) - len(post_state.dfs[sheet_index].index)

        return post_state, {
            **execution_data,
            'result': {
                'num_rows_dropped': num_rows_dropped
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            DropDuplicatesCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'keep')
            )   
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.transpiler.transpile_utils import get_column_header_as_transpiled_code
from mitosheet.state import State

class ColumnHeadersTransformCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, transformation: Any):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.transformation = transformation

        self.df_name = self.prev_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Column Headers Transform'
    
    def get_description_comment(self) -> str:
        if self.transformation['type'] == 'replace':
            return f"Replaced column headers in {self.df_name}"

        return f"Transformed headers in {self.df_name} to {self.transformation['type']}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = []
        if self.transformation['type'] == 'uppercase':
            code.append(f"{self.df_name}.columns = [col.upper() if isinstance(col, str) else col for col in {self.df_name}.columns]")
        elif self.transformation['type'] == 'lowercase':
            code.append(f"{self.df_name}.columns = [col.lower() if isinstance(col, str) else col for col in {self.df_name}.columns]")
        elif self.transformation['type'] == 'replace':
            old = get_column_header_as_transpiled_code(self.transformation['old'])
            new = get_column_header_as_transpiled_code(self.transformation['new'])
            code.append(f"{self.df_name}.columns = [col.replace({old}, {new}) if isinstance(col, str) else col for col in {self.df_name}.columns]")
        else:
            raise ValueError(f'Unknown transformation type: {self.transformation["type"]}')

        return code, []        

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, ColumnHeadersTransformCodeChunk):
            if _is_uppercase_or_lowercase_column_header_transform_code_chunk(self) and _is_uppercase_or_lowercase_column_header_transform_code_chunk(other_code_chunk):
                return other_code_chunk
            
        return None

def _is_uppercase_or_lowercase_column_header_transform_code_chunk(code_chunk: CodeChunk) -> bool:
        return isinstance(code_chunk, ColumnHeadersTransformCodeChunk) and (code_chunk.transformation['type'] == 'uppercase' or code_chunk.transformation['type'] == 'lowercase')



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
import os
from typing import Any, Dict, Optional
from mitosheet.types import SnowflakeCredentials, StepsManagerType

# The snowflake-connector-python package is only available in Python > 3.6 
# and is not distributed with the mitosheet package, so we make sure to 
# not assume that the import will succeed. 
try:
    import snowflake.connector
    SNOWFLAKE_CONNECTOR_IMPORTED = True
except ImportError:
    SNOWFLAKE_CONNECTOR_IMPORTED = False

# Global variable used to cache the snowflake credentials so that users
# only need to enter credentials once per kernel's lifespan. This global variable is accessed 
# by all mitosheets in the notebook! Caching the credentials also allows us to not pass 
# the snowflake credentials to the step performer, which ensures they don't get written into the analysis json.
__cached_snowflake_credentials: Optional[SnowflakeCredentials] = None

def get_cached_snowflake_credentials() -> Optional[SnowflakeCredentials]:
    global __cached_snowflake_credentials

    # If we can auth off of the env variables, then we do so and cache the credentials
    if SNOWFLAKE_CONNECTOR_IMPORTED \
        and 'SNOWFLAKE_ACCOUNT' in os.environ and 'SNOWFLAKE_USERNAME' in os.environ and 'SNOWFLAKE_PASSWORD' in os.environ \
        and __cached_snowflake_credentials is None:
        # Check if the snowflake credentials are valid
        exception = get_exception_from_snowflake_credentials(
            os.environ['SNOWFLAKE_USERNAME'],
            os.environ['SNOWFLAKE_PASSWORD'],
            os.environ['SNOWFLAKE_ACCOUNT'],
        )
        
        if exception is None:
            # If they are valid, then we cache them for the rest of the kernel's lifespan
            __cached_snowflake_credentials = {
                'type': 'success', # TODO: what should this be?
                'username': os.environ['SNOWFLAKE_USERNAME'],
                'password': os.environ['SNOWFLAKE_PASSWORD'],
                'account': os.environ['SNOWFLAKE_ACCOUNT'],
            }

    return __cached_snowflake_credentials

def get_exception_from_snowflake_credentials(username: str, password: str, account: str) -> Optional[Exception]:
        try:
            con = snowflake.connector.connect(
                    user=username,
                    password=password,
                    account=account,
            )
            con.close() #type: ignore
            return None
        except Exception as e:
            return e




def get_validate_snowflake_credentials(params: SnowflakeCredentials, steps_manager: StepsManagerType) -> Dict[str, Any]:
    """
    Takes Snowflake Credentials and validates them by creating a snowflake connection. If it succeeds, it stores the credentials
    as a global variable and returns a success object. If it fails, returns an error object.

    This API call __must__ be called before the user can perform a snowflake_import or getting available options and defaults for
    the snowflake query.
    """

    if not SNOWFLAKE_CONNECTOR_IMPORTED: 
        return {
            'type': 'error',    
            'error_message': 'snowflake-connector-python not accessible. Ensure it is installed.'
        }

    username = params['username']
    password = params['password']
    account = params['account']

    exception = get_exception_from_snowflake_credentials(username, password, account)

    if exception is not None:
        return {
            'type': 'error',    
            'error_message': f'{exception}'
        }

    # cache the snowflake credentials 
    global __cached_snowflake_credentials
    __cached_snowflake_credentials = params
        
    return {
        'type': 'success'
    }
                



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.melt_code_chunk import MeltCodeChunk

from mitosheet.state import DATAFRAME_SOURCE_MELTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID
from mitosheet.utils import get_first_unused_dataframe_name

class MeltStepPerformer(StepPerformer):
    """
    Allows you to melt.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'melt'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        id_var_column_ids: List[ColumnID] = get_param(params, 'id_var_column_ids')
        value_var_column_ids: List[ColumnID] = get_param(params, 'value_var_column_ids')

        id_vars = prev_state.column_ids.get_column_headers_by_ids(sheet_index, id_var_column_ids)
        value_vars = prev_state.column_ids.get_column_headers_by_ids(sheet_index, value_var_column_ids)

        # We remove all id vars from value vars, as in earlier versions of pandas overlap causes
        # errors that we don't want
        value_vars = list(filter(lambda value_var: value_var not in id_vars, value_vars))

        # If we're not including any headers that wouldn't be included by default, then we can 
        # just leave the value_vars empty
        include_value_vars = not set(prev_state.dfs[sheet_index].columns).difference(id_vars).issubset(value_vars)

        new_df_name = get_first_unused_dataframe_name(prev_state.df_names, f'{prev_state.df_names[sheet_index]}_unpivoted')
        execution_data = {
            'include_value_vars': include_value_vars,
            'new_df_name': new_df_name
        }

        return cls.execute_through_transpile(
            prev_state, 
            params, 
            execution_data, 
            {
                'df_source': DATAFRAME_SOURCE_MELTED,
                'new_df_names': [new_df_name],
                'overwrite': None
            }
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            MeltCodeChunk(
                prev_state,
                get_param(params, 'sheet_index'),
                get_param(params, 'id_var_column_ids'),
                get_param(params, 'value_var_column_ids'),
                get_param(execution_data if execution_data is not None else {}, 'include_value_vars'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_name'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
import random
import string
from copy import copy, deepcopy
from typing import Any, Callable, Collection, Dict, List, Optional, Set, Tuple, Union

import pandas as pd
from mitosheet.api.get_parameterizable_params import get_parameterizable_params_metadata
from mitosheet.api.get_path_contents import get_path_parts

from mitosheet.enterprise.mito_config import MitoConfig
from mitosheet.enterprise.telemetry.mito_log_uploader import MitoLogUploader
from mitosheet.experiments.experiment_utils import get_current_experiment
from mitosheet.step_performers.column_steps.set_column_formula import get_user_defined_sheet_function_objects
from mitosheet.step_performers.import_steps.dataframe_import import DataframeImportStepPerformer
from mitosheet.step_performers.import_steps.excel_range_import import ExcelRangeImportStepPerformer
from mitosheet.step_performers.user_defined_import import UserDefinedImportStepPerformer
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.preprocessing import PREPROCESS_STEP_PERFORMERS
from mitosheet.saved_analyses.save_utils import get_analysis_exists
from mitosheet.state import State
from mitosheet.step import Step
from mitosheet.step_performers import EVENT_TYPE_TO_STEP_PERFORMER
from mitosheet.step_performers.import_steps.excel_import import \
    ExcelImportStepPerformer
from mitosheet.step_performers.import_steps.simple_import import \
    SimpleImportStepPerformer
from mitosheet.step_performers.import_steps.snowflake_import import \
    SnowflakeImportStepPerformer
from mitosheet.transpiler.transpile import transpile
from mitosheet.transpiler.transpile_utils import get_default_code_options
from mitosheet.types import CodeOptions, MitoTheme, ParamMetadata
from mitosheet.updates import UPDATES
from mitosheet.user.utils import is_enterprise, is_running_test
from mitosheet.utils import NpEncoder, dfs_to_array_for_json, get_new_id, is_default_df_names
from mitosheet.step_performers.utils.user_defined_function_utils import get_user_defined_importers_for_frontend, get_user_defined_editors_for_frontend
from mitosheet.step_performers.utils.user_defined_function_utils import validate_and_wrap_sheet_functions, validate_user_defined_editors

def get_step_indexes_to_skip(step_list: List[Step]) -> Set[int]:
    """
    Given a list of steps, will collect all of the steps
    from this list that should be skipped.
    """
    step_indexes_to_skip: Set[int] = set()

    for step_index, step in enumerate(step_list):
        step_indexes_to_skip = step_indexes_to_skip.union(
            step.step_indexes_to_skip(step_list[:step_index])
        )

    return step_indexes_to_skip


def execute_step_list_from_index(
    step_list: List[Step], start_index: Optional[int]=None
) -> List[Step]:
    """
    Given a list of steps, and a specific index to start from, will assume that
    the step_list[start_index] is valid, and execute this list of steps from
    there. See find_last_valid_index to see how the start_index is determined.

    Notably, will _not_ execute the steps that are skipped in this list, which
    means that the returned step list will only have valid prev_state/post_states
    for the steps that are not skipped.

    If start_index is not given, will start from the initialize step.
    """

    # Make sure start index is not None
    if start_index is None:
        start_index = 0

    # Get the steps to skip, so that we can skip them
    step_indexes_to_skip = get_step_indexes_to_skip(step_list)

    # Get the steps that are valid, and the last valid step, so we can execute from there
    new_step_list = step_list[: start_index + 1]
    last_valid_step = step_list[start_index]

    for partial_index, step in enumerate(step_list[start_index + 1 :]):
        step_index = partial_index + start_index + 1
        # If we're skipping a step, add it to the new step list (since we don't
        # want to lose it), but don't reexecute it
        if step_index in step_indexes_to_skip:
            new_step_list.append(step)
            continue
            
        # Create a new step with the same params
        new_step = Step(step.step_type, step.step_id, step.params)

        # Set the previous state of the new step, and then update
        # what the last valid step is. Note that we find the actually
        # executed steps before passing them
        non_skipped_steps = [step for index, step in enumerate(new_step_list) if index not in step_indexes_to_skip]
        new_step.set_prev_state_and_execute(last_valid_step.final_defined_state, non_skipped_steps)
        last_valid_step = new_step

        new_step_list.append(new_step)

    return new_step_list


def get_modified_sheet_indexes(
    steps: List[Step], starting_step_index: int, ending_step_index: int
) -> Set[int]:
    """
    Returns a best guess for which sheets have been modified starting at
    starting_step_index and ending at (and including) ending_step_index.

    This is a best guess for caching reasons, and so may return sheets that
    have in fact not been modified. If a sheet index has been modified, it should
    always be returned.
    """
    # If only one step has been performed, we can calculate the modified sheet indexes,
    # otherwise we just say all of them (undo, replay might interact weird)
    if starting_step_index == ending_step_index - 1:
        step = steps[ending_step_index]
        modified_indexes = step.step_performer.get_modified_dataframe_indexes(step.params)

        # If the set is empty, then we modified everything
        if len(modified_indexes) == 0:
            modified_indexes = {j for j in range(len(step.dfs))}
        # If -1 is modified, then all new dataframes are modified, which
        # if nothing new was created, means there was a live updated event,
        # and so we should just take the last element
        elif -1 in modified_indexes:
            prev_step = steps[ending_step_index - 1]
            modified_indexes.remove(-1)

            if len(prev_step.dfs) != len(step.dfs):
                modified_indexes.update(
                    {j for j in range(len(prev_step.dfs), len(step.dfs))}
                )
            else:
                modified_indexes.add(len(step.dfs) - 1)
    else:
        modified_indexes = {i for i in range(len(steps[ending_step_index].dfs))}

    return modified_indexes


class StepsManager:
    """
    The StepsManager holds the list of the steps, and makes sure
    they are updated properly when processing new edit and update
    events.

    To see how this happens, consider the full loop of what happens
    when a new edit event is received.

    1.  The StepsManager receives this new event. It creates a new
        step of the right type, with the params from the new event.
        This is done in the handle_edit_event function.
    2.  Then, the StepsManager tries to run all the steps that exist
        including this new step _that are not currently skipped_.
        It does so with the execute_and_update_steps function, which heavily
        relies on the execute_step_list_from_index helper function above.
    3.  If the new step that is created is a step that does overwrite an
        existing step, then the execute_and_update_steps will detect
        this, and when running all the steps, will skip running the skipped step.

    In this way, the StepsManager keeps a list of all the steps that were ever
    valid. Note that some part of these steps is "append-only" in spirit, in
    that they keep the same step type and params.

    However, as some steps end up getting skipped, it may contain steps that
    are now invalid or out of date (e.g. an old filter step that got skipped).

    Furthermore, the steps objects themselves change, just the step type
    and parameters stay the same and are append-only.
    """

    def __init__(
            self, 
            args: Collection[Union[pd.DataFrame, str, None]], 
            mito_config: MitoConfig, 
            mito_log_uploader: Optional[MitoLogUploader]=None,
            analysis_to_replay: Optional[str]=None,
            import_folder: Optional[str]=None,
            user_defined_functions: Optional[List[Callable]]=None,
            user_defined_importers: Optional[List[Callable]]=None,
            user_defined_editors: Optional[List[Callable]]=None,
            code_options: Optional[CodeOptions]=None,
            theme: Optional[MitoTheme]=None,
        ):
        """
        When initalizing the StepsManager, we also do preprocessing
        of the arguments that were passed to the mitosheet.

        All preprocessing can be found in mitosheet/preprocessing, and each of
        the transformations are applied before the data is considered imported.
        """

        # We just randomly generate analysis names as a string of 10 letters
        self.analysis_name = 'id-' + ''.join(random.choice(string.ascii_lowercase) for _ in range(10))

        # We also save some data about the analysis the user wants to replay, if there
        # is such an analysis
        self.analysis_to_replay = analysis_to_replay
        self.analysis_to_replay_exists = get_analysis_exists(analysis_to_replay)

        # The import folder is the folder that users have the right to import files from. 
        # If this is set, then we should never let users view or access files that are not
        # inside this folder
        self.import_folder = import_folder

        # The args are a tuple of dataframes or strings, and we start by making them
        # into a list, and making copies of them for safe keeping
        self.original_args = [
            arg.copy(deep=True) if isinstance(arg, pd.DataFrame) else deepcopy(arg)
            for arg in args
        ]

        # Then, we go through the process of actually preprocessing the args
        # saving any data that we need to transpilate it later this
        self.preprocess_execution_data = {}
        df_names = None
        for preprocess_step_performers in PREPROCESS_STEP_PERFORMERS:
            args, df_names, execution_data = preprocess_step_performers.execute(args)
            self.preprocess_execution_data[
                preprocess_step_performers.preprocess_step_type()
            ] = execution_data       

        # We set the original_args_raw_strings. If we later have an args update, then these
        # are overwritten by the args update (and are actually correct). But since we don't 
        # always have an args update, this is the best we can do at this point in time. Notably, 
        # these are required to be set for transpiling arguments
        self.original_args_raw_strings: List[str] = []
        for index, arg in enumerate(self.original_args):
            if isinstance(arg, str):
                self.original_args_raw_strings.append(f'"{arg}"')
            elif df_names is not None and len(df_names) > index:
                self.original_args_raw_strings.append(df_names[index])
            else:
                self.original_args_raw_strings.append('')

        # Then, we check user defined functions. Check them for validity, and wrap them in the correct wrappers,
        self.user_defined_functions = validate_and_wrap_sheet_functions(user_defined_functions) 

        # We also do some checks for the user_defined_importers
        self.user_defined_importers = user_defined_importers
        if not is_running_test() and not is_enterprise() and self.user_defined_importers is not None and len(self.user_defined_importers) > 0:
            raise ValueError("importers are only supported in the enterprise version of Mito. See Mito plans https://www.trymito.io/plans")
        
        self.user_defined_editors = validate_user_defined_editors(user_defined_editors)
        if not is_running_test() and not is_enterprise() and self.user_defined_editors is not None and len(self.user_defined_editors) > 0:
            raise ValueError("editors are only supported in the enterprise version of Mito. See Mito plans https://www.trymito.io/plans")
        
        # The version of the public interface used by this analysis
        self.public_interface_version = 3

        # Then we initialize the analysis with just a simple initialize step
        self.steps_including_skipped: List[Step] = [
            Step(
                "initialize", "initialize", {}, None, 
                State(
                    args, 
                    self.public_interface_version,
                    df_names=df_names,
                    user_defined_functions=self.user_defined_functions, 
                    user_defined_importers=self.user_defined_importers,
                    user_defined_editors=self.user_defined_editors
                ), 
                {}
            )
        ]

        """
        To help with redo, we store a list of a list of the steps that 
        existed in the step manager before the user clicked undo or reset,
        along with the type of operation.

        This object contains two types of tuples:
        1. ('append', Step)
        2. ('reset', Step[])

        If you undo a single step, an ('append', Step) will be added onto this 
        list. If you then redo, this Step will be appended back onto the list. 

        If you clear or update_existing_imports, then a ('reset', Step[]) will
        be appended onto this list. If you then undo, this list will become the
        entire list of steps.
        """
        self.undone_step_list_store: List[Tuple[str, List[Step]]] = []

        # We display the state that exists after the curr_step_idx is applied,
        # which means you can never see before the initalize step
        self.curr_step_idx = 0

        # We also cache some of the sheet data in a form suitable to turn
        # into json, so that we can package it and send it to the front-end
        # faster and with less work. Make sure to cache the starting values
        # for the saved sheet data
        self.saved_sheet_data: List[Dict] = dfs_to_array_for_json(
            self.curr_step.final_defined_state,
            set(range(len(args))),
            [],
            self.curr_step.dfs,
            self.curr_step.df_names,
            self.curr_step.df_sources,
            self.curr_step.column_formulas,
            self.curr_step.column_filters,
            self.curr_step.column_ids,
            self.curr_step.df_formats,
        )
        self.last_step_index_we_wrote_sheet_json_on = 0

        # We store the number of update events that have been processed successfully,
        # which allows us to have some awareness about undos and redos in the front-end
        self.update_event_count = 0
        self.redo_count = 0
        self.undo_count = 0

        # This stores the number of times that the sheet renders, and we use it to detect
        # when we are on the first render of a sheet. This is very useful for making
        # sure we only update the state of the backend on the first render of a sheet
        # that corresponds to that backend
        self.render_count = 0

        # We store the experiment that is currently being run for this user
        self.experiment = get_current_experiment()

        # We store the mito_config variables here so that we can use them in the api
        self.mito_config = mito_config

        # Store the mito_log_uploader
        self.mito_log_uploader = mito_log_uploader

        # The options for the transpiled code. The user can optionally pass these 
        # in, but if they don't, we use the default options
        # We also do some checks for the user_defined_importers
        if not is_running_test() and not is_enterprise() and code_options is not None and len(code_options) > 0:
            raise ValueError("code_options are only supported in the enterprise version of Mito. See Mito plans https://www.trymito.io/plans")

        self.code_options: CodeOptions = get_default_code_options(self.analysis_name) if code_options is None else code_options

        self.theme = theme

    @property
    def curr_step(self) -> Step:
        """
        Returns the current step object as a property of the object,
        so reference it with self.curr_step
        """
        return self.steps_including_skipped[self.curr_step_idx]

    @property
    def dfs(self) -> List[pd.DataFrame]:
        return self.steps_including_skipped[self.curr_step_idx].dfs

    @property
    def sheet_data_json(self) -> str:
        """
        sheet_json contains a serialized representation of the data
        frames that is then fed into the Endo in the front-end.

        NOTE: we only display the _first_ 1,500 rows of the dataframe
        for speed reasons. This results in way less data getting
        passed around
        """
        modified_sheet_indexes = get_modified_sheet_indexes(
            self.steps_including_skipped, self.last_step_index_we_wrote_sheet_json_on, self.curr_step_idx
        )

        array = dfs_to_array_for_json(
            self.curr_step.final_defined_state,
            modified_sheet_indexes,
            self.saved_sheet_data,
            self.curr_step.dfs,
            self.curr_step.df_names,
            self.curr_step.df_sources,
            self.curr_step.column_formulas,
            self.curr_step.column_filters,
            self.curr_step.column_ids,
            self.curr_step.df_formats,
        )

        self.saved_sheet_data = array
        self.last_step_index_we_wrote_sheet_json_on = self.curr_step_idx

        return json.dumps(array, cls=NpEncoder)

    @property
    def analysis_data_json(self):
        return json.dumps(
            {
                "analysisName": self.analysis_name,
                "publicInterfaceVersion": self.public_interface_version,
                "analysisToReplay": {
                    'analysisName': self.analysis_to_replay,
                    'existsOnDisk': self.analysis_to_replay_exists,
                } if self.analysis_to_replay is not None else None,
                "code": self.code(),
                "stepSummaryList": self.step_summary_list,
                "currStepIdx": self.curr_step_idx,
                "graphDataDict": self.curr_step.graph_data_dict,
                'updateEventCount': self.update_event_count,
                'undoCount': self.undo_count,
                'redoCount': self.redo_count,
                'renderCount': self.render_count,
                'lastResult': self.curr_step.execution_data['result'] if 'result' in self.curr_step.execution_data else None,
                'experiment': self.experiment,
                'codeOptions': self.code_options,
                'userDefinedFunctions': get_user_defined_sheet_function_objects(self.curr_step.post_state),
                'userDefinedImporters': get_user_defined_importers_for_frontend(self.curr_step.post_state),
                'userDefinedEdits': get_user_defined_editors_for_frontend(self.curr_step.post_state),
                "importFolderData": {
                    'path': self.import_folder,
                    'pathParts': get_path_parts(self.import_folder)
                } if self.import_folder is not None else None,
                "theme": self.theme
            },
            cls=NpEncoder
        )

    @property
    def step_summary_list(self) -> List:
        """
        Returns a json list of step summaries, not including
        the skipped steps
        """
        step_summary_list = []
        step_indexes_to_skip = get_step_indexes_to_skip(self.steps_including_skipped)
        for index, step in enumerate(self.steps_including_skipped):
            if step.step_type == "initialize":
                step_summary_list.append(
                    {
                        "step_id": step.step_id,
                        "step_idx": 0,
                        "step_type": step.step_type,
                        "step_display_name": "Created a mitosheet",
                        "step_description": "Created a new mitosheet",
                        "params": step.params,
                        "result": step.execution_data.get('result', None) if step.execution_data else None
                    }
                )
                continue

            if index in step_indexes_to_skip:
                continue
            
            # NOTE: we cannot and should not optimize the code chunks here, as
            # rely on getting data out of them is to label the steps correctly
            code_chunks = step.step_performer.transpile(
                step.prev_state, # type: ignore
                step.params,
                step.execution_data,
            )

            step_summary_list.append(
                {
                    "step_id": step.step_id,
                    "step_idx": index,
                    "step_type": step.step_type,
                    "step_display_name": code_chunks[0].get_display_name(),
                    "step_description": code_chunks[0].get_description_comment().strip().replace('\n', '\n# '),
                    "params": step.params,
                    "result": step.execution_data.get('result', None) if step.execution_data else None
                }
            )

        return step_summary_list
    
    def code(self) -> List[str]:
        return transpile(self, optimize=True)
    
    @property
    def fully_parameterized_function(self) -> str:
        """
        Returns the fully parameterized function string. This is used for
        cases where we want to get the function string regardless of the 
        code options the user provided. 
        """
        return '\n'.join(transpile(
            self,
            add_comments=False,
            optimize=True,
            code_options_override={
                'import_custom_python_code': True,
                'as_function': True,
                'call_function': False,
                'function_name': self.code_options.get('function_name', 'automate'),
                'function_params': 'all'
            }
        ))
    
    @property
    def param_metadata(self) -> List[ParamMetadata]:
        return get_parameterizable_params_metadata(self)

    def handle_edit_event(self, edit_event: Dict[str, Any]) -> None:
        """
        Updates the widget state with a new step that was created
        by the edit_event. Each edit event creates one new step.

        If there is an error in the creation of the new step, this
        function will not create the new invalid step.
        """

        # NOTE: We ignore any edit if we are in a historical state, for now. This is a result
        # of the fact that we don't allow previous editing currently
        if self.curr_step_idx != len(self.steps_including_skipped) - 1:
            return

        step_performer = EVENT_TYPE_TO_STEP_PERFORMER[edit_event["type"]]

        # First, we add the public interface to the params, as we might need it for any step
        edit_event["params"]['public_interface_version'] = self.public_interface_version

        # Then, we make a new step
        new_step = Step(
            step_performer.step_type(), edit_event["step_id"], edit_event["params"]
        )

        new_steps = self.steps_including_skipped + [new_step]

        self.execute_and_update_steps(new_steps)

        # If we add a new step, then we clear the last_undone_list_store, as
        # you cannot redo something after you make a new edit
        self.undone_step_list_store = []

        # NOTE: we also check here if we're receiving an edit event, and we have not
        # yet read in the non-default dataframe names. We have to log this here, rather
        # than in the .sheet call (or elsewhere) because the dataframe names are read
        # in after the sheet is rendered. Thus, we just check after the first edit event
        # (e.g. when there are two steps) - if we still have default dataframe names, this
        # is an error. Note we make this a distinct log from when the args update itself
        # fails so that we can check if we really do get to this state
        if len(self.steps_including_skipped) == 2 and is_default_df_names(self.curr_step.df_names): # NOTE: two means we have done at least one edit.
            log('args_update_remains_failed')

    def handle_update_event(self, update_event: Dict[str, Any]) -> None:
        """
        Handles any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names
        or some existing steps).
        """

        for update in UPDATES:
            if update_event["type"] == update["event_type"]:
                # Get the params for this event
                params = {key: value for key, value in update_event['params'].items() if key in update['params']}  # type: ignore
                # Actually execute this event
                update["execute"](self, **params)  # type: ignore
                # Update the number of update events we record occuring
                self.update_event_count += 1
                # And then return
                return

        raise Exception(f"{update_event} is not an update event!")

    def find_last_valid_index(self, new_steps: List[Step]) -> int:
        """
        Given the new_steps, this function performs some logic to figure
        out what the last valid index in the steps is (that execution can
        then start from).
        """

        # Currently, we only remove steps in an undo
        if len(new_steps) < len(self.steps_including_skipped):
            # If we are removing steps, then we figure out what skipped steps
            # we are losing, and run from right before where we are no longer
            # skipped steps
            no_longer_skipped_indexes: Set[int] = set()
            for step_index, removed_step in enumerate(self.steps_including_skipped[len(new_steps) :]):
                previous_steps = self.steps_including_skipped[: len(new_steps) + step_index]
                no_longer_skipped_indexes = no_longer_skipped_indexes.union(
                    removed_step.step_indexes_to_skip(previous_steps)
                )

            last_valid_index = (
                min(no_longer_skipped_indexes.union({len(new_steps)})) - 1
            )
        else:
            # Otherwise, if we're adding steps, we figure out which skipped steps
            # we're adding, and run from right before the oldest new skipped step

            # Collect anything that is newly skipped
            newly_skipped_indexes: Set[int] = set()
            for step_index, new_step in enumerate(new_steps[len(self.steps_including_skipped) :]):
                previous_steps = new_steps[: len(self.steps_including_skipped) + step_index]
                newly_skipped_indexes = newly_skipped_indexes.union(
                    new_step.step_indexes_to_skip(previous_steps)
                )

            # The last valid index is the minimum of the newly skipped things - 1
            # or the last valid step (if nothing is skipped)
            last_valid_index = min(newly_skipped_indexes.union({len(self.steps_including_skipped)})) - 1

        # Make sure that this step isn't itself skipped, and decrement until it is not
        all_skipped_indexes = get_step_indexes_to_skip(new_steps)
        while last_valid_index in all_skipped_indexes:
            last_valid_index -= 1

        # Make sure that this index is positive -- it always should be!
        if last_valid_index < 0:
            return 0

        return last_valid_index

    def execute_undo(self) -> None:
        """
        This function attempts to undo the most recent step, and if there
        is no most recent step, does nothing.

        It also handles the special case where the last action was clearing
        the analysis, in which case we have the undo actually reset the
        entire analysis.
        """

        self.undo_count += 1

        # When a user's most recent action is a clear analysis or update_existing_imports, then the undone_step_list_store
        # will end in an item that says ('reset', [...]).
        # In this case, if they press undo right after clearing, then we assume they probably
        # want to undo the clear, aka to redo all those steps
        if len(self.undone_step_list_store) > 0:
            if self.undone_step_list_store[-1][0] == "reset" or self.undone_step_list_store[-1][0] == "undo_to_step_index":
                return self.execute_redo()

        # Otherwise, we just undo the most recent step that the user has created
        # if they have created any steps
        if len(self.steps_including_skipped) == 1:
            return

        new_steps = copy(self.steps_including_skipped)
        undone_step = new_steps.pop()

        self.execute_and_update_steps(new_steps)

        # If this works, then let's add this step to the undo list!
        self.undone_step_list_store.append(("append", [undone_step]))

    def execute_redo(self) -> None:
        """
        Executes a redo, which reapplies the most recently undone
        steps if they exist.

        This will not error if there is nothing to redo, it will
        just return.
        """
        self.redo_count += 1

        if len(self.undone_step_list_store) == 0:
            return

        (undo_or_clear, step_list) = self.undone_step_list_store[-1]
        if undo_or_clear == "append":
            # If it's an undo, just apply onto the end
            new_steps = copy(self.steps_including_skipped)
            new_steps.extend(step_list)
            self.execute_and_update_steps(new_steps)

        elif undo_or_clear == "reset":
            new_steps = step_list
            # Note: since we're breaking the invariant that the steps don't
            # move order, we have to execute from the very start
            self.execute_and_update_steps(new_steps, last_valid_index=0)

        elif undo_or_clear == "undo_to_step_index":
            new_steps = step_list
            self.execute_and_update_steps(new_steps)

        # Remove the item we just redid from the undone_step_list_store, so
        # that we don't redo it again
        self.undone_step_list_store.pop()

    def execute_clear(self) -> None:
        """
        A clear update, which removes all steps in the analysis
        that are not imports.
        """
        if len(self.steps_including_skipped) == 1:
            return

        # Keep only the initalize step and any import steps
        new_steps = [
            step
            for step in self.steps_including_skipped
            if (
                step.step_type == "initialize"
                or step.step_type == SimpleImportStepPerformer.step_type()
                or step.step_type == ExcelImportStepPerformer.step_type()
                or step.step_type == DataframeImportStepPerformer.step_type()
                or step.step_type == SnowflakeImportStepPerformer.step_type()
                or step.step_type == ExcelRangeImportStepPerformer.step_type()
                or step.step_type == UserDefinedImportStepPerformer.step_type()
            )
        ]

        old_steps = copy(self.steps_including_skipped)

        # We need to set the last_valid_index to 0 as we are changing
        # the step order
        self.execute_and_update_steps(new_steps, last_valid_index=0)

        self.undone_step_list_store.append(("reset", old_steps))

    def execute_undo_to_step_index(self, step_idx: int) -> None:
        """
        A execute_undo_to_step_index update, removes all steps in the analysis
        that come after the provided step_idx
        """
        if len(self.steps_including_skipped) == 1:
            return

        old_steps = copy(self.steps_including_skipped)

        new_steps = self.steps_including_skipped[:step_idx + 1]

        # We need to set the last_valid_index to 0 as we are changing
        # the step order
        self.execute_and_update_steps(new_steps, last_valid_index=0)

        self.undone_step_list_store.append(("undo_to_step_index", old_steps))


    def execute_and_update_steps(
        self, new_steps: List[Step], last_valid_index: Optional[int] = None
    ) -> None:
        """
        Given a list of new_steps, runs them from the last valid index,
        based on what is skipped in these new_steps.

        If these new steps result in valid execution, then we set them
        to the step that the step manager stores.

        If last_valid_index is passed, then this will be used instead of
        the last_valid_index that could be calculated by the step list.

        So, pass a last_valid_index if you're changing the order of the steps
        in the new_steps array. Otherwise, the step manager can calculate
        the last valid index without help.
        """
        if last_valid_index is None:
            last_valid_index = self.find_last_valid_index(new_steps)

        final_steps = execute_step_list_from_index(
            new_steps, start_index=last_valid_index
        )
        self.steps_including_skipped = final_steps
        self.curr_step_idx = len(self.steps_including_skipped) - 1

    def execute_steps_data(self, new_steps_data: Optional[List[Dict[str, Any]]] = None) -> None:
        """
        Given steps data (e.g. from a saved analysis), will turn
        this data  into steps and try to run them. If any of them
        fail, will take none of the new steps
        """
        new_steps = copy(self.steps_including_skipped)
        if new_steps_data:
            for step_data in new_steps_data:
                new_step = Step(
                    step_data["step_type"], get_new_id(), step_data["params"]
                )

                new_steps.append(new_step)

        self.execute_and_update_steps(new_steps)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
This file contains helpful functions and classes for testing operations.
"""

from copy import deepcopy
import json
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd
from numpy import number

from mitosheet.code_chunks.code_chunk_utils import get_code_chunks
from mitosheet.mito_backend import MitoBackend, get_mito_backend
from mitosheet.step_performers.graph_steps.plotly_express_graphs import (
    DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT, DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT,
    DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT)
from mitosheet.step_performers.pivot import PCT_NO_OP
from mitosheet.transpiler.transpile import transpile
from mitosheet.transpiler.transpile_utils import (
    get_column_header_list_as_transpiled_code, get_column_header_as_transpiled_code, get_globals_for_exec)
from mitosheet.types import (CodeOptions, ColumnHeader, ColumnHeaderWithFilter,
                             ColumnHeaderWithPivotTransform, ColumnID,
                             ColumnIDWithFilter, ColumnIDWithPivotTransform,
                             DataframeFormat, FormulaAppliedToType, GraphID,
                             MultiLevelColumnHeader, Filter, FilterGroup, OperatorType)
from mitosheet.utils import NpEncoder, dfs_to_array_for_json, get_new_id


def check_transpiled_code_after_call(func):
    @wraps(func)
    def wrapper(*args, **kw):
        result = func(*args, **kw)
        check_dataframes_equal(args[0])
        return result
    return wrapper


def check_dataframes_equal(test_wrapper: "MitoWidgetTestWrapper") -> None:
    """
    Tests that the dataframes in the widget state container equal
    to those that are the result of the executed code. 

    Also makes sure that the sheet json that is saved is the correct
    sheet json that is used in the sheet
    """

    # The only dataframes we want to define apriori are the dataframes that
    # were passed directly to the mito widget
    original_args = {
        arg_name: arg.copy(deep=True) if isinstance(arg, pd.DataFrame) else deepcopy(arg) for arg, arg_name in 
        zip(
            test_wrapper.mito_backend.steps_manager.original_args,
            test_wrapper.mito_backend.steps_manager.steps_including_skipped[0].df_names
        )
    }
    final_dfs = {
        df_name: df.copy(deep=True) for df, df_name in 
        zip(
            test_wrapper.mito_backend.steps_manager.curr_step.dfs,
            test_wrapper.mito_backend.steps_manager.curr_step.df_names
        )
    }

    # Then, construct code that is just the code we expect, except at the end
    # it compares the dataframe to the final dataframe we expect
    def check_final_dataframe(df_name, df):
        assert final_dfs[df_name].equals(df)

    code = "\n".join(
        test_wrapper.transpiled_code +
        [
            f'check_final_dataframe(\'{df_name}\', {df_name})'
            for df_name in test_wrapper.mito_backend.steps_manager.curr_step.df_names
        ]
    )

    final_state = test_wrapper.mito_backend.steps_manager.curr_step.final_defined_state

    try:
        exec(code, 
            {
                'check_final_dataframe': check_final_dataframe,
                **get_globals_for_exec(final_state, final_state.public_interface_version),
            }, 
            original_args
        )
    except:
        from mitosheet.errors import get_recent_traceback
        print("\n\nError executing code")
        print(f'Unoptimized code chunks: {test_wrapper.unoptimized_code_chunks}')
        print(f'Optimized code chunks: {test_wrapper.optimized_code_chunks}')
        print(get_recent_traceback())
        print("Code:")
        print(code)
        raise

    # We then check that the sheet data json that is saved by the widget, which 
    # notably uses caching, does not get incorrectly cached and is written correctly
    sheet_data_json = test_wrapper.mito_backend.get_shared_state_variables()['sheet_data_json']
    assert sheet_data_json == json.dumps(dfs_to_array_for_json(
        test_wrapper.mito_backend.steps_manager.curr_step.final_defined_state, 
        set(i for i in range(len(test_wrapper.mito_backend.steps_manager.curr_step.dfs))),
        [],
        test_wrapper.mito_backend.steps_manager.curr_step.dfs,
        test_wrapper.mito_backend.steps_manager.curr_step.df_names,
        test_wrapper.mito_backend.steps_manager.curr_step.df_sources,
        test_wrapper.mito_backend.steps_manager.curr_step.column_formulas,
        test_wrapper.mito_backend.steps_manager.curr_step.column_filters,
        test_wrapper.mito_backend.steps_manager.curr_step.column_ids,
        test_wrapper.mito_backend.steps_manager.curr_step.df_formats
    ), cls=NpEncoder)


class MitoWidgetTestWrapper:
    """
    This class adds some simple wrapper functions onto the MitoWidget 
    to make interacting with it easier for testing purposes.

    It allows you to create just the backend piece of Mito, create columns,
    set formulas, and get values to check the result.
    """

    def __init__(self, mito_backend: MitoBackend):
        self.mito_backend = mito_backend

    @property
    def transpiled_code(self):
        # NOTE: we don't add comments to this testing functionality, so that 
        # we don't have to change tests if we update comments
        return transpile(self.mito_backend.steps_manager, add_comments=False)

    @property
    def transpiled_code_with_comments(self):
        return transpile(self.mito_backend.steps_manager, add_comments=True)
    
    @property
    def unoptimized_code_chunks(self):
        return get_code_chunks(self.mito_backend.steps_manager.steps_including_skipped, optimize=False)

    @property
    def optimized_code_chunks(self):
        # NOTE: we don't add comments to this testing functionality, so that 
        # we don't have to change tests if we update comments
        return get_code_chunks(self.mito_backend.steps_manager.steps_including_skipped, optimize=True)

    @property
    def curr_step_idx(self):
        return self.mito_backend.steps_manager.curr_step_idx
    
    @property
    def steps_including_skipped(self):
        return self.mito_backend.steps_manager.steps_including_skipped

    @property
    def curr_step(self):
        return self.mito_backend.steps_manager.curr_step
    
    @property
    def dfs(self):
        return self.mito_backend.steps_manager.dfs

    @property
    def df_names(self):
        return self.mito_backend.steps_manager.curr_step.df_names

    @property
    def df_formats(self):
        return self.mito_backend.steps_manager.curr_step.df_formats
    
    @property
    def sheet_data_json(self):
        return self.mito_backend.steps_manager.sheet_data_json

    @property
    def analysis_data_json(self):
        return self.mito_backend.steps_manager.analysis_data_json

    @check_transpiled_code_after_call
    def add_column(self, sheet_index: int, column_header: str, column_header_index: int=-1) -> bool:
        """
        Adds a column.
        """

        return self.mito_backend.receive_message({
            'event': 'edit_event',
            'id': get_new_id(),
            'type': 'add_column_edit',
            'step_id': get_new_id(),
            'params': {
                'sheet_index': sheet_index,
                'column_header': column_header,
                'column_header_index': column_header_index
            }
        })
    
    @check_transpiled_code_after_call
    def set_formula(
            self, 
            formula: str, 
            sheet_index: int,
            column_header: str, 
            add_column: bool=False,
            formula_label: Optional[Any]=None,
            index_labels: Optional[List[Any]]=None
        ) -> bool:
        """
        Sets the given column to have formula, and optionally
        adds the column if it does not already exist.
        """
        if add_column:
            self.add_column(sheet_index, column_header)

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        if formula_label is None:
            formula_label = self.mito_backend.steps_manager.dfs[sheet_index].index[0]

        index_labels_formula_is_applied_to: FormulaAppliedToType
        if index_labels is None:
            index_labels_formula_is_applied_to = {'type': 'entire_column'}
        else:
            index_labels_formula_is_applied_to = {'type': 'specific_index_labels', 'index_labels': index_labels}

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'set_column_formula_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'formula_label': formula_label,
                    'index_labels_formula_is_applied_to': index_labels_formula_is_applied_to,
                    'new_formula': formula,
                }
            }
        )

    @check_transpiled_code_after_call
    def merge_sheets(
            self, 
            how: str,
            sheet_index_one: int, 
            sheet_index_two: int, 
            merge_key_columns: List[Tuple[ColumnHeader, ColumnHeader]], 
            selected_columns_one: List[ColumnHeader],
            selected_columns_two: List[ColumnHeader],
            destination_sheet_index: Optional[int]=None,
        ) -> bool:

        merge_key_column_ids = list(map(lambda x: [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index_one, x[0]),
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index_two, x[1]),
        ], merge_key_columns))

        selected_column_ids_one = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index_one, column_header)
            for column_header in selected_columns_one
        ]
        selected_column_ids_two = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index_two, column_header)
            for column_header in selected_columns_two
        ]

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'merge_edit',
                'step_id': get_new_id(),
                'params': {
                    'how': how,
                    'sheet_index_one': sheet_index_one,
                    'sheet_index_two': sheet_index_two,
                    'merge_key_column_ids': merge_key_column_ids,
                    'selected_column_ids_one': selected_column_ids_one,
                    'selected_column_ids_two': selected_column_ids_two,
                    'destination_sheet_index': destination_sheet_index
                }
            }
        )
    
    @check_transpiled_code_after_call
    def concat_sheets(
            self, 
            join: str,
            ignore_index: bool,
            sheet_indexes: int
        ) -> bool:

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'concat_edit',
                'step_id': get_new_id(),
                'params': {
                    'join': join,
                    'ignore_index': ignore_index,
                    'sheet_indexes': sheet_indexes
                }
            }
        )
    
    @check_transpiled_code_after_call
    def fill_na(
            self, 
            sheet_index: int,
            column_headers: List[ColumnHeader],
            fill_method: Any
        ) -> bool:

        column_ids = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index, column_header)
            for column_header in column_headers
        ]

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'fill_na_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_ids': column_ids,
                    'fill_method': fill_method
                }
            }
        )

    
    @check_transpiled_code_after_call
    def delete_row(
            self, 
            sheet_index: int,
            labels: List[Union[int, str]],
        ) -> bool:        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'delete_row_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'labels': labels,
                }
            }
        )
    

    @check_transpiled_code_after_call
    def promote_row_to_header(
            self, 
            sheet_index: int,
            index: Any,
        ) -> bool:
        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'promote_row_to_header_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'index': index,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def transpose(
            self, 
            sheet_index: int,
        ) -> bool:

        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'transpose_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def melt(
            self, 
            sheet_index: int,
            id_var_column_headers: List[ColumnHeader],
            value_var_column_headers: List[ColumnHeader],
        ) -> bool:

        id_var_column_ids = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index, column_header)
            for column_header in id_var_column_headers
        ]
        value_var_column_ids = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index, column_header)
            for column_header in value_var_column_headers
        ]

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'melt_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'id_var_column_ids': id_var_column_ids,
                    'value_var_column_ids': value_var_column_ids,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def one_hot_encoding(
            self, 
            sheet_index: int,
            column_header: ColumnHeader,
        ) -> bool:

        column_id =self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )


        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'one_hot_encoding_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def set_dataframe_format(
            self, 
            sheet_index: int,
            df_format: Any,
        ) -> bool:

        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'set_dataframe_format_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'df_format': df_format,
                    
                }
            }
        )
    

    # NOTE: We do not automatically check the generated code, as the variables are not defined
    # in this context, which cases these tests to always fail. I can't think of a good way around
    # this, so we're just gonna skip this for now
    def dataframe_import(
            self, 
            df_names: List[str],
        ) -> bool:

        
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'df_names': df_names,
                }
            }
        )
    

    @check_transpiled_code_after_call
    def snowflake_import(
            self, 
            table_loc_and_warehouse: Any,
            query_params: Any,
        ) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'snowflake_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'table_loc_and_warehouse': table_loc_and_warehouse,
                    'query_params': query_params,
                }
            }
        )

    @check_transpiled_code_after_call
    def excel_range_import(
            self, 
            file_path: str,
            sheet: Dict[str, Union[str, int]],
            range_imports: Any,
            convert_csv_to_xlsx: bool,
        ) -> bool:

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'excel_range_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'file_path': file_path,
                    'sheet': sheet,
                    'range_imports': range_imports,
                    'convert_csv_to_xlsx': convert_csv_to_xlsx
                }
            }
        )
    
    # NOTE: we don't automatically run the code for testing reasons, so that we can
    # test the step performer to make sure that it works properly without the generated
    # code overwriting the files
    def export_to_file(
            self, 
            type: str,
            sheet_indexes: List[int],
            file_name: str,
            export_formatting: bool=False
        ) -> bool:

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'export_to_file_edit',
                'step_id': get_new_id(),
                'params': {
                    'type': type,
                    'sheet_indexes': sheet_indexes,
                    'file_name': file_name,
                    'export_formatting': export_formatting
                }
            }
        )

    @check_transpiled_code_after_call
    def reset_index(
            self, 
            sheet_index: int,
            drop: bool,
        ) -> bool:

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'reset_index_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'drop': drop,
                }
            }
        )
    

    @check_transpiled_code_after_call
    def ai_transformation(
            self, 
            user_input: str,
            prompt_version: str,
            prompt: str,
            completion: str,
            edited_completion: str,
        ) -> bool:

        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'ai_transformation_edit',
                'step_id': get_new_id(),
                'params': {
                    'user_input': user_input,
                    'prompt_version': prompt_version,
                    'prompt': prompt,
                    'completion': completion,
                    'edited_completion': edited_completion,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def column_headers_transform(
            self, 
            sheet_index: int,
            transformation: Any,
        ) -> bool:

        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'column_headers_transform_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'transformation': transformation,
                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def user_defined_import(
            self, 
            importer: str,
            importer_params: Dict[str, str]
        ) -> bool:


        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'user_defined_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'importer': importer,
                    'importer_params': importer_params
                }
            }
        )
    

    @check_transpiled_code_after_call
    def replace(
            self, 
            sheet_index: int,
            column_ids: List[ColumnID],
            search_value: str,
            replace_value: str,
        ) -> bool:

        

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'replace_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_ids': column_ids,
                    'search_value': search_value,
                    'replace_value': replace_value,                    
                }
            }
        )
    

    @check_transpiled_code_after_call
    def user_defined_edit(
            self, 
            edit_name: str,
            edit_params: Any,
        ) -> bool:

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'user_defined_edit_edit',
                'step_id': get_new_id(),
                'params': {
                    'edit_name': edit_name,
                    'edit_params': edit_params,
                }
            }
        )
    
# AUTOGENERATED LINE: TEST (DO NOT DELETE)

    @check_transpiled_code_after_call
    def split_text_to_columns(
            self, 
            sheet_index: int,
            column_header: ColumnHeader,
            delimiters: List[str],
            new_column_header_suffix: str
        ) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.get_column_id_by_header(sheet_index, column_header)

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'split_text_to_columns_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'delimiters': delimiters,
                    'new_column_header_suffix': new_column_header_suffix
                }
            }
        )

    @check_transpiled_code_after_call
    def drop_duplicates(
            self, 
            sheet_index: int, 
            column_headers: List[ColumnHeader], 
            keep: str,
        ) -> bool:

        column_ids = [
            self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(sheet_index, column_header)
            for column_header in column_headers
        ]

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'drop_duplicates_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_ids': column_ids,
                    'keep': keep,
                }
            }
        )

    @check_transpiled_code_after_call
    def pivot_sheet(
            self, 
            sheet_index: int, 
            # For convenience, you can use this testing API to either pass just column headers, or optionally
            # column headers with transforms attached to them. This function turns them
            # into the correct format before passing them to the backend
            pivot_rows: Union[List[ColumnHeader], List[ColumnHeaderWithPivotTransform]],
            pivot_columns: Union[List[ColumnHeader], List[ColumnHeaderWithPivotTransform]],
            values: Dict[ColumnHeader, List[str]],
            pivot_filters: Optional[List[ColumnHeaderWithFilter]]=None,
            flatten_column_headers: bool=True,
            destination_sheet_index: Optional[int]=None,
            step_id: Optional[str]=None
        ) -> bool:

        get_column_id_by_header = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header

        rows_ids_with_transforms: List[ColumnIDWithPivotTransform] = []
        if len(pivot_rows) > 0 and isinstance(pivot_rows[0], str):
            rows_ids_with_transforms = [{
                'column_id': get_column_id_by_header(sheet_index, column_header), # type: ignore
                'transformation': PCT_NO_OP
            } for column_header in pivot_rows]
        elif len(pivot_rows) > 0:
            rows_ids_with_transforms = [{
                'column_id': get_column_id_by_header(sheet_index, chwpt['column_header']), # type: ignore
                'transformation': chwpt['transformation'] # type: ignore
            } for chwpt in pivot_rows]
        
        column_ids_with_transforms: List[ColumnIDWithPivotTransform] = []
        if len(pivot_columns) > 0 and isinstance(pivot_columns[0], str):
            column_ids_with_transforms = [{
                'column_id': get_column_id_by_header(sheet_index, column_header), # type: ignore
                'transformation': PCT_NO_OP
            } for column_header in pivot_columns]
        elif len(pivot_columns) > 0:
            column_ids_with_transforms = [{
                'column_id': get_column_id_by_header(sheet_index, chwpt['column_header']), # type: ignore
                'transformation': chwpt['transformation'] # type: ignore
            } for chwpt in pivot_columns]


        values_column_ids_map = {
            get_column_id_by_header(sheet_index, column_header): value
            for column_header, value in values.items()
        }

        pivot_filters_ids: List[ColumnIDWithFilter] = [
            {'column_id': get_column_id_by_header(sheet_index, pf['column_header']), 'filter': pf['filter']} 
            for pf in pivot_filters
        ] if pivot_filters is not None else []

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'pivot_edit',
                'step_id': get_new_id() if step_id is None else step_id,
                'params': {
                    'sheet_index': sheet_index,
                    'pivot_rows_column_ids_with_transforms': rows_ids_with_transforms,
                    'pivot_columns_column_ids_with_transforms': column_ids_with_transforms,
                    'values_column_ids_map': values_column_ids_map,
                    'destination_sheet_index': destination_sheet_index,
                    'pivot_filters': pivot_filters_ids,
                    'flatten_column_headers': flatten_column_headers
                }
            }
        )

    @check_transpiled_code_after_call
    def filter(
            self, 
            sheet_index: int, 
            column_header: ColumnHeader,
            operator: OperatorType,
            condition: str, 
            value: Any
        ) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'filter_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'operator': operator,
                    'filters': [{
                        'condition': condition,
                        'value': value
                    }]
                }
            }
        )

    
    @check_transpiled_code_after_call
    def filters(
            self, 
            sheet_index: int, 
            column_header: ColumnHeader,
            operator: OperatorType,
            filters: List[Union[Filter, FilterGroup]]
        ) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'filter_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'operator': operator,
                    'filters': filters
                }
            }
        )
    
    @check_transpiled_code_after_call
    def sort(
            self, 
            sheet_index: int, 
            column_header: ColumnHeader,
            sort_direction: str,
            step_id: Optional[str]=None
        ) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'sort_edit',
                'step_id': get_new_id() if step_id is None else step_id,
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'sort_direction': sort_direction
                }
            }
        )

    @check_transpiled_code_after_call
    def reorder_column(
            self, 
            sheet_index: int, 
            column_header: ColumnHeader, 
            new_column_index: int
        ) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'reorder_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'new_column_index': new_column_index
                }
            }
        )

    @check_transpiled_code_after_call
    def rename_column(self, sheet_index: int, old_column_header: ColumnHeader, new_column_header: ColumnHeader, level: Optional[int]=None) -> bool:

        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            old_column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'rename_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'new_column_header': new_column_header,
                    'level': level
                }
            }
        )

    @check_transpiled_code_after_call
    def delete_columns(self, sheet_index: int, column_headers: List[ColumnHeader]) -> bool:
        column_ids = [self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header 
        ) for column_header in column_headers]

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'delete_column_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_ids': column_ids,
                }
            }
        )

    @check_transpiled_code_after_call
    def change_column_dtype(self, sheet_index: int, column_headers: List[ColumnHeader], new_dtype: str) -> bool:

        column_ids = self.mito_backend.steps_manager.curr_step.column_ids.get_column_ids_by_headers(
            sheet_index,
            column_headers
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'change_column_dtype_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_ids': column_ids,
                    'new_dtype': new_dtype
                }
            }
        )

    @check_transpiled_code_after_call
    def simple_import(
        self, 
        file_names: List[str], 
        delimeters: Optional[List[str]]=None, 
        encodings: Optional[List[str]]=None, 
        decimals: Optional[List[str]]=None, 
        skiprows: Optional[List[int]]=None,
        error_bad_lines: Optional[List[bool]]=None
    ) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'simple_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'file_names': file_names,
                    'delimeters': delimeters,
                    'encodings': encodings,
                    'decimals': decimals,
                    'skiprows': skiprows,
                    'error_bad_lines': error_bad_lines,
                }
            }
        )

    @check_transpiled_code_after_call
    def excel_import(self, file_name: str, sheet_names: List[str], has_headers: bool, skiprows: int, decimal: Optional[str]=None) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'excel_import_edit',
                'step_id': get_new_id(),
                'params': {
                    'file_name': file_name,
                    'sheet_names': sheet_names,
                    'has_headers': has_headers,
                    'skiprows': skiprows,
                    'decimal': decimal
                }   
            }
        )

    @check_transpiled_code_after_call
    def bulk_old_rename(self, move_to_deprecated_id_algorithm: bool=False) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'old_rename_only_use_this_in_testing',
                'step_id': get_new_id(),
                'params': {
                    'move_to_deprecated_id_algorithm': move_to_deprecated_id_algorithm
                }
            }
        )

    @check_transpiled_code_after_call
    def undo(self) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'undo',
                'params': {},
            }
        )
    
    @check_transpiled_code_after_call
    def argsUpdate(self, args: List[str]) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'args_update',
                'params': {
                    'args': args
                },
            }
        )

    @check_transpiled_code_after_call
    def redo(self) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'redo',
                'params': {},
            }
        )

    @check_transpiled_code_after_call
    def clear(self) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'clear',
                'params': {},
            }
        )
    

    def save_analysis(self, analysis_name: str) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'save_analysis_update',
                'params': {
                    'analysis_name': analysis_name
                },
            }
        )


    @check_transpiled_code_after_call
    def delete_dataframe(self, sheet_index: int) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_delete_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index
                }
            }
        )

    @check_transpiled_code_after_call
    def duplicate_dataframe(self, sheet_index: int) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_duplicate_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                }
            }
        )

    @check_transpiled_code_after_call
    def rename_dataframe(self, sheet_index: int, new_dataframe_name: str) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'dataframe_rename_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'new_dataframe_name': new_dataframe_name
                }
            }
        )

    @check_transpiled_code_after_call
    def set_cell_value(self, sheet_index: int, column_header: ColumnHeader, row_index: int, new_value: Any) -> bool:
        column_id = self.mito_backend.steps_manager.curr_step.column_ids.get_column_id_by_header(
            sheet_index,
            column_header
        )

        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'set_cell_value_edit',
                'step_id': get_new_id(),
                'params': {
                    'sheet_index': sheet_index,
                    'column_id': column_id,
                    'row_index': row_index,
                    'new_value': str(new_value)
                }
            }
        )

    @check_transpiled_code_after_call
    def replay_analysis(self, analysis_name: str, args: Optional[List[str]]=None, step_import_data_list_to_overwrite: Optional[List[Dict[str, Any]]]=None) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'replay_analysis_update',
                'params': {
                    'analysis_name': analysis_name,
                    'args': args if args is not None else [],
                    'step_import_data_list_to_overwrite': step_import_data_list_to_overwrite if step_import_data_list_to_overwrite is not None else []
                },
            }
        )
    
    @check_transpiled_code_after_call
    def code_options_update(self, code_options: CodeOptions) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'code_options_update',
                'params': {
                    'code_options': code_options
                },
            }
        )
    
    # Some code options allow you to turn off the actually calling of a function, and
    # in this case, we don't want to check the transpiled code, as it will always fail
    def code_options_update_no_check_transpiled(self, code_options: CodeOptions) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'code_options_update',
                'params': {
                    'code_options': code_options
                },
            }
        )

    @check_transpiled_code_after_call
    def checkout_step_by_idx(self, index: int) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'checkout_step_by_idx_update',
                'params': {
                    'step_idx': index,
                },
            }
        )

    @check_transpiled_code_after_call
    def undo_to_step_index(self, index: int) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'undo_to_step_index_update',
                'params': {
                    'step_idx': index,
                },
            }
        )
    
    
    @check_transpiled_code_after_call
    def checklist_update(self, checklist_id: str, completed_items: List[str], clear_other_items: bool) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'id': get_new_id(),
                'type': 'checklist_update',
                'params': {
                    'checklist_id': checklist_id,
                    'completed_items': completed_items,
                    'clear_other_items': clear_other_items
                },
            }
        )


    def generate_graph(
        self, 
        graph_id: str,
        graph_type: str, 
        sheet_index: number,
        safety_filter_turned_on_by_user: bool,
        xAxisColumnIDs: List[ColumnID],
        yAxisColumnIDs: List[ColumnID],
        height: str,
        width: str,
        color: Optional[ColumnID]=None,
        facet_col_column_id: Optional[ColumnID]=None,
        facet_row_column_id: Optional[ColumnID]=None,
        facet_col_wrap: Optional[int]=None,
        facet_col_spacing: Optional[float]=None,
        facet_row_spacing: Optional[float]=None,
        title_title: Optional[str]=None,
        title_visible: bool=True,
        title_font_color: str=DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT,
        xaxis_title: Optional[str]=None,
        xaxis_visible: bool=True,
        xaxis_title_font_color: str=DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT,
        xaxis_type: Optional[str]=None,
        xaxis_showgrid: bool=True,
        xaxis_gridwidth: Optional[number]=None,
        xaxis_rangeslider_visible: bool=True,
        yaxis_title: Optional[str]=None,
        yaxis_visible: bool=True,
        yaxis_title_font_color: str=DO_NOT_CHANGE_TITLE_FONT_COLOR_DEFAULT,
        yaxis_type: Optional[str]=None,
        yaxis_showgrid: bool=True,
        yaxis_gridwidth: Optional[number]=None,
        showlegend: bool=True,
        legend_title_text: Optional[str]=None,
        legend_orientation: Optional[str]='v',
        legend_x: Optional[number]=None,
        legend_y: Optional[number]=None,
        step_id: Optional[str]=None,
        paper_bgcolor: str=DO_NOT_CHANGE_PAPER_BGCOLOR_DEFAULT,
        plot_bgcolor: str=DO_NOT_CHANGE_PLOT_BGCOLOR_DEFAULT,
        barmode: Optional[str]=None,
        barnorm: Optional[str]=None,
        histnorm: Optional[str]=None,
        histfunc: Optional[str]=None,
        nbins: Optional[number]=None,
        line_shape: Optional[str]=None,
        points: Optional[str]=None,

    ) -> bool:

        params: Any = {
            'graph_id': graph_id,
            'graph_preprocessing': {
                'safety_filter_turned_on_by_user': safety_filter_turned_on_by_user
            },
            'graph_creation': {
                'graph_type': graph_type,
                'sheet_index': sheet_index,
                'x_axis_column_ids': xAxisColumnIDs,
                'y_axis_column_ids': yAxisColumnIDs,
                'color': color,
                'facet_col_column_id': facet_col_column_id,
                'facet_row_column_id': facet_row_column_id,
                'histfunc': histfunc,
                'histnorm': histnorm,
                'line_shape': line_shape,
                'points': points
            },
            'graph_styling': {
                'title': {
                    'title': title_title,
                    'visible': title_visible,
                    'title_font_color': title_font_color
                },
                'xaxis': {
                    'title': xaxis_title,
                    'visible': xaxis_visible,
                    'title_font_color': xaxis_title_font_color,
                    'type': xaxis_type,
                    'showgrid': xaxis_showgrid,
                    'rangeslider': {
                        'visible': xaxis_rangeslider_visible
                    }
                },
                'yaxis': {
                    'title': yaxis_title,
                    'visible': yaxis_visible,
                    'title_font_color': yaxis_title_font_color,
                    'type': yaxis_type,
                    'showgrid': yaxis_showgrid,
                },
                'showlegend': showlegend,
                'legend': {
                    'title': {
                        'text': legend_title_text
                    },
                    'orientation': legend_orientation,
                },
                'barmode': barmode,
                'barnorm': barnorm,
                'paper_bgcolor': paper_bgcolor,
                'plot_bgcolor': plot_bgcolor,
            },
            'graph_rendering': {
                'height': height,
                'width': width
            }
        }

        # We add these params because when they the backend castst them to a number. The backend doesn't handle the None case because
        # none params are filtered out. 
        # Instead of handing the None case specifically for the tests, we keep our code simple by mocking the filtering out of None.
        if facet_col_wrap is not None:
            params['graph_creation']['facet_col_wrap'] = facet_col_wrap

        if facet_col_spacing is not None:
            params['graph_creation']['facet_col_spacing'] = facet_col_spacing

        if facet_row_spacing is not None:
            params['graph_creation']['facet_row_spacing'] = facet_row_spacing   

        if xaxis_gridwidth is not None:
            params['graph_styling']['xaxis']['gridwidth'] = xaxis_gridwidth  

        if yaxis_gridwidth is not None:
            params['graph_styling']['yaxis']['gridwidth'] = yaxis_gridwidth 

        if legend_x is not None:
            params['graph_styling']['legend']['x'] = legend_x 

        if legend_y is not None:
            params['graph_styling']['legend']['y'] = legend_y      

        if nbins is not None:
            params['graph_creation']['nbins'] = nbins

        return self.mito_backend.receive_message( 
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'graph_edit',
                'step_id': get_new_id() if step_id is None else step_id,
                'params': params
            }
        )

    def delete_graph(self, graph_id: GraphID) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'graph_delete_edit',
                'step_id': get_new_id(),
                'params': {
                    'graph_id': graph_id,
                }
            }
        )

    def duplicate_graph(self, old_graph_id: GraphID, new_graph_id: GraphID) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'graph_duplicate_edit',
                'step_id': get_new_id(),
                'params': {
                    'old_graph_id': old_graph_id,
                    'new_graph_id': new_graph_id
                }
            }
        )


    def rename_graph(self, graph_id: GraphID, new_graph_tab_name: str) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'edit_event',
                'id': get_new_id(),
                'type': 'graph_rename_edit',
                'step_id': get_new_id(),
                'params': {
                    'graph_id': graph_id,
                    'new_graph_tab_name': new_graph_tab_name
                }
            }
        )

    def update_existing_imports(self, updated_import_objs: List[Dict[str, Any]]) -> bool:
        return self.mito_backend.receive_message(
            {
                'event': 'update_event',
                'type': 'update_existing_import_update',
                'id': get_new_id(),
                'params': {
                    'updated_step_import_data_list': updated_import_objs
                }
            }
        )

    def get_formula(self, sheet_index: int, column_header: ColumnHeader) -> str:
        """
        Gets the formula for a given column. Returns an empty
        string if nothing exists.
        """
        column_id = self.mito_backend.steps_manager.curr_step.get_column_id_by_header(
            sheet_index, column_header
        )
        if column_id not in self.mito_backend.steps_manager.curr_step.column_formulas[sheet_index]:
            return ''
        return self.mito_backend.steps_manager.curr_step.column_formulas[sheet_index][column_id]


    def get_value(self, sheet_index: int, column_header: ColumnHeader, row: int) -> Any:
        """
        Returns a value in a given dataframe at the given
        index in a column. NOTE: the row is 1 indexed!

        Errors if the value does not exist
        """
        return self.mito_backend.steps_manager.curr_step.dfs[sheet_index].at[row - 1, column_header]

    def get_column(self, sheet_index: int, column_header: ColumnHeader, as_list: bool) -> Union[pd.Series, List]:
        """
        Returns a series object of the given column, or a list if
        as_list is True. 

        Errors if the column does not exist. 
        """
        if as_list:
            return self.mito_backend.steps_manager.dfs[sheet_index][column_header].tolist()
        return self.mito_backend.steps_manager.dfs[sheet_index][column_header]

    def get_graph_data(self, graph_id: str) -> Dict[str, Dict[str, Any]]: 
        """
        Returns the graph_data object 
        """
        if graph_id in self.mito_backend.steps_manager.curr_step.final_defined_state.graph_data_dict.keys():
            return self.mito_backend.steps_manager.curr_step.final_defined_state.graph_data_dict[graph_id]
        else:
            return {}

    def get_graph_type(self, graph_id: str) -> str:
        """
        Returns the graph type 
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["graph_type"]
        return ''

    def get_graph_sheet_index(self, graph_id: str) -> int:
        """
        Returns the graph sheet index 
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["sheet_index"]
        return -1

    def get_graph_axis_column_ids(self, graph_id: str, axis: str) -> List[str]:
        """
        Returns the graph axis column ids for either the x or y axis
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            if axis == 'x':
                return graph_data["graphParams"]["graphCreation"]["x_axis_column_ids"]
            if axis == 'y':
                return graph_data["graphParams"]["graphCreation"]["y_axis_column_ids"]
        return []

    def get_graph_color(self, graph_id: str) -> ColumnID:
        """
        Returns the graph color column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["color"]
        return ''

    def get_graph_facet_col_column_id(self, graph_id: str) -> ColumnID:
        """
        Returns the graph facet col column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["facet_col_column_id"]
        return ''

    def get_graph_facet_row_column_id(self, graph_id: str) -> ColumnID:
        """
        Returns the graph facet row column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["facet_row_column_id"]
        return ''

    def get_graph_facet_col_wrap(self, graph_id: str) -> int:
        """
        Returns the graph facet row column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["facet_col_wrap"]
        return -1

    def get_graph_facet_col_spacing(self, graph_id: str) -> float:
        """
        Returns the graph facet row column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["facet_col_spacing"]
        return -1.0

    def get_graph_facet_row_spacing(self, graph_id: str) -> float:
        """
        Returns the graph facet row column id
        """
        graph_data = self.get_graph_data(graph_id)
        if bool(graph_data):
            return graph_data["graphParams"]["graphCreation"]["facet_row_spacing"]
        return -1.0
    
    def get_is_graph_output_none(self, graph_id: str) -> bool:
        """
        Returns true if all of the graphOuput is does not exist.
        """
        return "graphOutput" not in self.get_graph_data(graph_id)

    def get_graph_styling_params(self, graph_id: str) -> Dict[str, Optional[Union[str, bool]]]:
        """
        Returns the object that stores all the graph styling params, so that 
        we can easily make sure the structure is correct
        """
        graph_data = self.get_graph_data(graph_id)
        return graph_data["graphParams"]["graphStyling"]

    def get_dataframe_format(self, sheet_index: int) -> DataframeFormat: 
        """
        Returns the DataframeFormat object for a specific sheet
        """
        return self.mito_backend.steps_manager.curr_step.final_defined_state.df_formats[sheet_index]
        

def create_mito_wrapper_with_data(sheet_one_A_data: List[Any], sheet_two_A_data: Optional[List[Any]]=None) -> MitoWidgetTestWrapper:
    """
    Returns a MitoWidgetTestWrapper instance wrapped around a MitoWidget
    that contains just a column A, containing sheet_one_A_data.
    
    If sheet_two_A_data is defined, then also creates a second dataframe
    with column A defined as this as well.
    """
    dfs = [pd.DataFrame(data={'A': sheet_one_A_data})]

    if sheet_two_A_data is not None:
        dfs.append(pd.DataFrame(data={'A': sheet_two_A_data}))

    mito_backend = get_mito_backend(*dfs)
    return MitoWidgetTestWrapper(mito_backend)


def create_mito_wrapper(
        *args: Union[pd.DataFrame, str], 
        arg_names: Optional[List[str]]=None,
        sheet_functions: Optional[List[Callable]]=None,
        importers: Optional[List[Callable]]=None,
        editors: Optional[List[Callable]]=None,
    ) -> MitoWidgetTestWrapper:
    """
    Creates a MitoWidgetTestWrapper with a mito instance with the given
    data frames.
    """
    mito_backend = get_mito_backend(*args, user_defined_functions=sheet_functions, user_defined_importers=importers, user_defined_editors=editors)
    test_wrapper =  MitoWidgetTestWrapper(mito_backend)

    if arg_names is not None:
        test_wrapper.argsUpdate(arg_names)

    # Check generated code for preprocessing is correct
    check_dataframes_equal(test_wrapper)

    return test_wrapper

def make_multi_index_header_df(data: Dict[Union[str, int], List[Any]], column_headers: List[ColumnHeader], index: Optional[List[Any]]=None) -> pd.DataFrame:
    """
    A helper function that allows you to easily create a multi-index
    header dataframe. 

    Simply pass the data you want in the dataframe in a dictonary, and then 
    a list of the column headers you want for each column in the data. 

    At least one of the column headers should be a tuple, so that this
    creates a multi-index header dataframe.

    So: make_multi_index_header_df({0: ['a'], 1: ['b']}, ['header', ('header', 'other')]) 
    will return a dataframe with the headers [('header', ''), ('header', 'other)].
    """
    df = pd.DataFrame(data=data)
    max_length = 0
    for column_header in column_headers:
        if isinstance(column_header, tuple):
            max_length = max(max_length, len(column_header))
    
    final_column_headers: List[ColumnHeader] = []
    for column_header in column_headers:
        if isinstance(column_header, tuple) or isinstance(column_header, list):
            final_column_headers.append(column_header)
        else:
            final_column_header: MultiLevelColumnHeader = [column_header] + ['' for _ in range(max_length - 1)]
            final_column_headers.append(final_column_header)

    df.columns = pd.MultiIndex.from_tuples(final_column_headers)
    if index is not None:
        df.index = index
    return df

def get_dataframe_generation_code(df: pd.DataFrame) -> str:
    """
    Given a dataframe like:

        date (year)      value sum
    0         2000             1.0
    1         2001             NaN

    Will return the string:
    pd.DataFrame({'date (year)': [2000, 2001], 'value sum: [1.0, np.NaN]})

    This is useful when you have a dataframe you want to create at runtime and then put into a test.
    """
    OPEN_BRACKET = "{"
    CLOSE_BRACKET = "}"

    data = ", ".join([f"{get_column_header_as_transpiled_code(column_header)}: {get_column_header_list_as_transpiled_code(df[column_header].to_list())}" for column_header in df.columns])
    return f'pd.DataFrame({OPEN_BRACKET}{data}{CLOSE_BRACKET})'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
For going to a timedelta series.
"""

from typing import Any, Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.is_type_utils import is_bool_dtype, is_datetime_dtype, is_number_dtype, is_string_dtype, is_timedelta_dtype


def to_timedelta_series(
        unknown_object: Any,
        on_uncastable_arg_element: Any=('default', np.NaN), #Union[Literal['error'], Tuple[Literal['default'], any]]
    ) -> pd.Series:
    """
    Converts the given object to a timedelta series. Note that on_uncastable_arg_element
    is irrelevant here.
    """

    # If it is not a series, we put it in a series, and get the type again
    if not isinstance(unknown_object, pd.Series):
        unknown_object = pd.Series([unknown_object])

    column_dtype = str(unknown_object.dtype)
    if is_bool_dtype(column_dtype):
        return None
    elif is_datetime_dtype(column_dtype):
        return None
    elif is_timedelta_dtype(column_dtype):
        return unknown_object
    elif is_number_dtype(column_dtype):
        return pd.to_timedelta(
            unknown_object, 
            unit='s',
            errors='coerce'
        )
    elif is_string_dtype(column_dtype):
        return pd.to_timedelta(
            unknown_object, 
            unit='s',
            errors='coerce'
        )
    else:
        return None



from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from mitosheet.is_type_utils import is_bool_dtype, is_datetime_dtype, is_float_dtype, is_int_dtype, is_string_dtype, is_timedelta_dtype

from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.types.bool import cast_to_bool
from mitosheet.public.v3.types.datetime import cast_series_to_datetime, cast_to_datetime
from mitosheet.public.v3.types.float import cast_to_float
from mitosheet.public.v3.types.int import cast_to_int
from mitosheet.public.v3.types.number import cast_to_number
from mitosheet.public.v3.types.str import cast_to_string
from mitosheet.public.v3.types.timedelta import cast_to_timedelta
from mitosheet.types import PrimitiveTypeName

SERIES_CONVERSION_FUNCTIONS: Dict[PrimitiveTypeName, Optional[Callable[[pd.Series], pd.Series]]] = {
    'str': None,
    'int': None,
    'float': None,
    'number': None,
    'bool': None,
    'datetime': cast_series_to_datetime,
    'timedelta': None,
}

ELEMENT_CONVERSION_FUNCTIONS: Dict[PrimitiveTypeName, Callable[[Any], Optional[Any]]] = {
    'str': cast_to_string,
    'int': cast_to_int,
    'float': cast_to_float,
    'number': cast_to_number,
    'bool': cast_to_bool,
    'datetime': cast_to_datetime,
    'timedelta': cast_to_timedelta,
}

def is_primitive_value(value: Any) -> bool:
    return isinstance(value, str) or \
        isinstance(value, int) or \
        isinstance(value, float) or \
        isinstance(value, bool) or \
        isinstance(value, datetime) or \
        isinstance(value, timedelta)
        # TODO: nan here?

def get_primitive_type_name_from_primitive_value(primitive_value: Any) -> PrimitiveTypeName:
    if isinstance(primitive_value, str):
        return 'str'
    elif isinstance(primitive_value, int):
        return 'int'
    elif isinstance(primitive_value, float):
        return 'float'
    elif isinstance(primitive_value, bool):
        return 'bool'
    elif isinstance(primitive_value, datetime):
        return 'datetime'
    elif isinstance(primitive_value, timedelta):
        return 'timedelta'
    else:
        return 'str'
    
def get_primitive_type_name_from_series(series: pd.Series) -> PrimitiveTypeName:
    dtype = str(series.dtype)
    if is_string_dtype(dtype):
        return 'str'
    elif is_int_dtype(dtype):
        return 'int'
    elif is_float_dtype(dtype):
        return 'float'
    elif is_bool_dtype(dtype):
        return 'bool'
    elif is_datetime_dtype(dtype):
        return 'datetime'
    elif is_timedelta_dtype(dtype):
        return 'timedelta'
    else:
        return 'str'


def get_arg_cast_to_type(
        target_primitive_type_name: PrimitiveTypeName, 
        arg: Any,
        primitive_types_to_ignore: Optional[List[PrimitiveTypeName]]=None
    ) -> Any:

    if primitive_types_to_ignore is None:
        primitive_types_to_ignore = []

    series_conversion_function_without_skip = SERIES_CONVERSION_FUNCTIONS[target_primitive_type_name]
    element_conversion_function_without_skip = ELEMENT_CONVERSION_FUNCTIONS[target_primitive_type_name]

    # Wrapper that make sure we skip any primitive types we want to ignore
    def element_conversion_function(arg: Any) -> Any:
        arg_type_name = get_primitive_type_name_from_primitive_value(arg)
        if arg_type_name in primitive_types_to_ignore: # type: ignore
            return arg

        return element_conversion_function_without_skip(arg)

    if series_conversion_function_without_skip is not None:
        def series_conversion_function(arg: Any) -> Any:
            arg_type_name = get_primitive_type_name_from_series(arg)
            if arg_type_name in primitive_types_to_ignore: # type: ignore
                return arg

            return series_conversion_function_without_skip(arg) # type: ignore
    else:
        series_conversion_function = None # type: ignore

    if is_primitive_value(arg):
        return element_conversion_function(arg)
    
    elif isinstance(arg, pd.Series):
        if series_conversion_function is not None:
            return series_conversion_function(arg)

        return arg.apply(element_conversion_function)

    elif isinstance(arg, pd.DataFrame):

        if series_conversion_function is not None:
            return arg.apply(lambda c: series_conversion_function(c))

        return arg.apply(lambda c: c.apply(element_conversion_function))

    elif isinstance(arg, RollingRange):
        obj = arg.obj

        if series_conversion_function is not None:
            new_obj = obj.apply(lambda c: series_conversion_function(c))
        else:
            new_obj = obj.apply(lambda c: c.apply(element_conversion_function))
            
        return RollingRange(new_obj, arg.window, arg.offset)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple, Union
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.is_type_utils import is_datetime_dtype, is_string_dtype, is_timedelta_dtype
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import get_column_header_list_as_transpiled_code, get_column_header_as_transpiled_code, get_param_dict_as_code
from mitosheet.types import ColumnHeader, ColumnID
from mitosheet.user.utils import get_pandas_version
from mitosheet.utils import is_prev_version

def get_split_param_dict() -> Dict[str, Optional[Union[bool, int]]]:
    # Create a dictionary of the params needed to configure the .split function for
    # split text to columns

    split_param_dict: Dict[str, Optional[Union[bool, int]]] = {
        'n': -1,
        'expand': True, 

    }

    # Regex was added to the split function on pandas 1.4.0. When the delimiter_string is . 
    # and regex is true (the default), it splits on every character. Setting regex=None:
    # If None and pat length is 1, treats pat as a literal string => . won't match on everything
    # If None and pat length is not 1, treats pat as a regular expression => splitting on ',|\t' will split on all commas and tabs
    pandas_version = get_pandas_version()
    if not is_prev_version(pandas_version, '1.4.0'):
        split_param_dict['regex'] = None 

    return split_param_dict


class SplitTextToColumnsCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_id: ColumnID, delimiters: List[str], new_column_headers: List[ColumnHeader]):
        super().__init__(prev_state)
        self.sheet_index: int = sheet_index
        self.column_id: ColumnID = column_id
        self.delimiters: List[str] = delimiters
        self.new_column_headers: List[ColumnHeader] = new_column_headers

        self.df_name = self.prev_state.df_names[self.sheet_index]
        self.column_header = self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        self.column_dtype = str(self.prev_state.dfs[self.sheet_index][self.column_header].dtype) # type: ignore


    def get_display_name(self) -> str:
        return 'Split text to columns'
    
    def get_description_comment(self) -> str:
        delimiters_string = (', ').join(map(lambda x: f'"{x}"', self.delimiters))
        return f'Split {self.column_header} on {delimiters_string}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        delimiter_string = repr('|'.join(self.delimiters))
        
        column_header = self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        transpiled_column_header = get_column_header_as_transpiled_code(column_header)
        new_transpiled_column_headers = get_column_header_list_as_transpiled_code(self.new_column_headers)
        column_idx = self.prev_state.column_ids.get_column_ids(self.sheet_index).index(self.column_id)

        # Split column
        if is_string_dtype(self.column_dtype):
            string_conversion = ''
        elif is_datetime_dtype(self.column_dtype):
            string_conversion = ".dt.strftime('%Y-%m-%d %X')"
        elif is_timedelta_dtype(self.column_dtype):
            string_conversion = ".apply(lambda x: str(x))"
        else:
            string_conversion = ".astype('str')"

        split_param_dict = get_split_param_dict()
        split_param_code = get_param_dict_as_code(split_param_dict, as_single_line=True)
            
        split_column_line = f'{self.df_name}[{new_transpiled_column_headers}] = {self.df_name}[{transpiled_column_header}]{string_conversion}.str.split({delimiter_string}, {split_param_code})'

        # Reorder columns 
        reorder_columns_line = f'{self.df_name} = {self.df_name}[{self.df_name}.columns[:{column_idx + 1}].tolist() + {new_transpiled_column_headers} + {self.df_name}.columns[{column_idx + 1}:-{len(self.new_column_headers)}].tolist()]'

        return [split_column_line, reorder_columns_line], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests for the state class
"""
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, DATAFRAME_SOURCE_PASSED, State
import pandas as pd

def test_state_can_add_df_to_end():
    df = pd.DataFrame({'A': [123]})
    state = State([df], 3)
    state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED)
    for key, value in state.__dict__.items():
        if isinstance(value, list) and key != 'user_defined_functions' and key != 'user_defined_importers' and key != 'user_defined_editors':
            assert len(value) == 2
    
    assert state.df_sources == [DATAFRAME_SOURCE_PASSED, DATAFRAME_SOURCE_IMPORTED]

def test_state_can_add_df_to_middle():
    df = pd.DataFrame({'A': [123]})
    state = State([df], 3)
    state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED, overwrite={'sheet_index_to_overwrite': 0, 'attempt_to_save_filter_metadata': False})
    for key, value in state.__dict__.items() :
        if isinstance(value, list) and key != 'user_defined_functions' and key != 'user_defined_importers' and key != 'user_defined_editors':
            assert len(value) == 1
    
    assert state.df_sources == [DATAFRAME_SOURCE_IMPORTED]




import pytest
from mitosheet.tests.decorators import requires_dash
from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito

@requires_dash
def test_cannot_create_dash_spreadsheet_component_outside_dash():
    with pytest.raises(Exception):
        Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'})
    
@requires_dash
def test_cannot_create_dash_spreadsheet_component_without_activating():
    from dash import Input, Output, State, Dash, html, dcc
    app = Dash(__name__)

    with pytest.raises(Exception):
        Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'})
    


@requires_dash
def test_mito_callback_component_callable_with_no_mito():
    from dash import Input, Output, State, Dash, html, dcc

    assert callable(mito_callback)

    # Make a dash app
    app = Dash(__name__)
    activate_mito(app)

    app.layout = html.Div([
        dcc.Input(id='input', value='initial value'),
        html.Div(id='output'),
        dcc.Store(id='state'),
    ])
    
    @mito_callback(
        Output('output', 'value'),
        Input('input', 'value'),
        State('state', 'value'),
    )
    def func(input_value, state_value):
        return input_value

@requires_dash
def test_mito_callback_component_callable_with_spreadsheet_as_input():
    from dash import Input, Output, State, Dash, html, dcc

    # Make a dash app
    app = Dash(__name__)
    activate_mito(app)

    app.layout = html.Div([
        dcc.Input(id='input', value='initial value'),
        html.Div(id='output'),
        dcc.Store(id='state'),
        Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'})
    ])
    
    @mito_callback(
        Output('output', 'value'),
        Input('input', 'value'),
        Input({'type': 'spreadsheet', 'id': 'sheet'}, 'mito_spreadsheet_result'),
        State('state', 'value'),
    )
    def func(input_value, spreadsheet_data, state_value):
        return input_value
    

@requires_dash
def test_mito_callback_component_callable_with_spreadsheet_in_state():
    from dash import Input, Output, State, Dash, html, dcc

    # Make a dash app
    app = Dash(__name__)
    activate_mito(app)

    app.layout = html.Div([
        dcc.Input(id='input', value='initial value'),
        html.Div(id='output'),
        dcc.Store(id='state'),
        Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'})
    ])
    
    @mito_callback(
        Output('output', 'value'),
        Input('input', 'value'),
        State({'type': 'spreadsheet', 'id': 'sheet'}, 'mito_spreadsheet_result'),
        State('state', 'value'),
    )
    def func(input_value, spreadsheet_data, state_value):
            return input_value
    
@requires_dash
def test_can_pass_track_selection():
    from dash import Input, Output, State, Dash, html, dcc

     # Make a dash app
    app = Dash(__name__)
    activate_mito(app)

    app.layout = html.Div([
        Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'}, track_selection=False)
    ])


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Rolls back to a specific step, by index
"""

from mitosheet.types import StepsManagerType


UNDO_TO_STEP_INDEX_UPDATE_EVENT = 'undo_to_step_index_update'
UNDO_TO_STEP_INDEX_UPDATE_PARAMS = [
    'step_idx'
]

def execute_following_steps_update(
        steps_manager: StepsManagerType,
        step_idx: int
    ) -> None:
    """
    Checks out a specific step by index
    """
    steps_manager.execute_undo_to_step_index(step_idx)

UNDO_TO_STEP_INDEX_UPDATE = {
    'event_type': UNDO_TO_STEP_INDEX_UPDATE_EVENT,
    'params': UNDO_TO_STEP_INDEX_UPDATE_PARAMS,
    'execute': execute_following_steps_update
}


from abc import ABC, abstractmethod
from typing import Any, Dict, Collection, List, Optional, Tuple

from mitosheet.types import StepsManagerType


class PreprocessStepPerformer(ABC, object):
    """
    The abstract base class for a preprocess step, which 
    executes at the start of the analysis on the arguments
    passed to the mitosheet.sheet call
    """

    @classmethod
    @abstractmethod
    def preprocess_step_version(cls) -> int:
        """
        Returns the version of the step. Changes when the parameters
        of the step change.
        """
        pass

    @classmethod
    @abstractmethod
    def preprocess_step_type(cls) -> str:
        """
        The name of the step used internally. If you change this, you must upgrade
        the step and bump the version.
        """
        pass

    @classmethod
    @abstractmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[List[str]], Optional[Dict[str, Any]]]:
        """
        Execute always returns the new list of arguments, as well as the names of these arguements, as well as execution_data
        for this preprocess step.
        """
        pass

    @classmethod
    @abstractmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> Tuple[List[str], List[str]]:
        """
        Returns a list of the Python code lines that corresponds to this 
        preprocess step being executed
        """
        pass



# V2 is the same as V1, but it has a new module, for testing reasons
from mitosheet.public.v1 import (
    sheet_functions, utils, flatten_column_header, deduplicate_column_headers,
    to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format,
    FUNCTIONS
)
from mitosheet.public.v1.sheet_functions import *
from mitosheet.public.v1 import register_analysis
from mitosheet.public.v2.excel_utils import get_table_range, get_read_excel_params_from_range, get_table_range, convert_csv_file_to_xlsx_file
import pandas as pd


"""
Mito aims to provide incredibly easy extension points. For now, Mito allows you to 
type a function variable with a ColumnHeader, and this will generate a UI for selecting
a column from the nearest preceding dataframe.
"""
from typing import Any
## Extension Types We Support 

# NOTE: the other approach here would be making ColumnHeader parameteric on the name of the dataframe
# arg. This would allow us to support multiple dataframes in a single function with potentially more 
# flexibility, but it feels needlessly complicated for now.
ColumnHeader = Any


# Function to encode a date into a license string
import base64
import datetime


def encode_date_to_license(date: datetime.date) -> str:
    date_str = date.strftime("%Y-%m-%d")
    encoded_date = base64.b64encode(date_str.encode()).decode()
    license_key = f"LICENSE-{encoded_date}-MITO-ENTERPRISE"
    return license_key

# Function to decode a license string back into a date
def decode_license_to_date(license_key: str) -> datetime.date:
    encoded_date = license_key.split("-")[1]    
    decoded_date_str = base64.b64decode(encoded_date).decode()    
    decoded_date = datetime.datetime.strptime(decoded_date_str, "%Y-%m-%d").date()
    return decoded_date


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains handlers for the Mito API
"""
from queue import Queue
from threading import Thread
from time import perf_counter
from typing import Any, Callable, Dict, NoReturn, Optional

from mitosheet.api.get_ai_completion import get_ai_completion
from mitosheet.api.get_available_snowflake_options_and_defaults import \
    get_available_snowflake_options_and_defaults
from mitosheet.api.get_code_snippets import get_code_snippets
from mitosheet.api.get_column_describe import get_column_describe
from mitosheet.api.get_column_summary_graph import get_column_summary_graph
from mitosheet.api.get_csv_files_metadata import get_csv_files_metadata
from mitosheet.api.get_dataframe_as_csv import get_dataframe_as_csv
from mitosheet.api.get_dataframe_as_excel import get_dataframe_as_excel
from mitosheet.api.get_defined_df_names import get_defined_df_names
from mitosheet.api.get_excel_file_metadata import get_excel_file_metadata
from mitosheet.api.get_imported_files_and_dataframes_from_analysis_name import \
    get_imported_files_and_dataframes_from_analysis_name
from mitosheet.api.get_imported_files_and_dataframes_from_current_steps import \
    get_imported_files_and_dataframes_from_current_steps
from mitosheet.api.get_parameterizable_params import get_parameterizable_params
from mitosheet.api.get_params import get_params
from mitosheet.api.get_path_contents import get_path_contents
from mitosheet.api.get_path_join import get_path_join
from mitosheet.api.get_pr_url_of_new_pr import get_pr_url_of_new_pr
from mitosheet.api.get_render_count import get_render_count
from mitosheet.api.get_search_matches import get_search_matches
from mitosheet.api.get_split_text_to_columns_preview import \
    get_split_text_to_columns_preview
from mitosheet.api.get_test_imports import get_test_imports
from mitosheet.api.get_unique_value_counts import get_unique_value_counts
from mitosheet.api.get_validate_snowflake_credentials import \
    get_validate_snowflake_credentials
from mitosheet.steps_manager import StepsManager
# AUTOGENERATED LINE: API.PY IMPORT (DO NOT DELETE)
from mitosheet.telemetry.telemetry_utils import log_event_processed
from mitosheet.types import MitoWidgetType
from mitosheet.user.location import is_dash, is_jupyterlite, is_streamlit

# As the column summary statistics tab does three calls, we defaulted to this max
MAX_QUEUED_API_CALLS = 3

# NOTE: BE CAREFUL WITH THIS. When in development mode, you can set it to False
# so the API calls are handled in the main thread, to make printing easy.
# In newer versions of JupyterLab, to see these print statements:
# View > Show Log Console > in the console set Log Level to Debug
THREADED = True


def get_api_should_be_threaded() -> bool:

    # JupyterLite does not support multiple threads, so we have to set it equal
    # to False in this case
    if is_jupyterlite():
        return False

    # Dashboarding solutions like Streamlit and Dash do their own thing with threads, so we
    # don't create a thread in that case either

    if is_streamlit():
        return False

    if is_dash():
        return False
    
    return True



class API:
    """
    The API provides a wrapper around a thread that responds to API calls.

    Some notes:
    -   We allow at most MAX_QUEUED_API_CALLS API calls to be in the queue, which practically
        Stops a backlog of calls from building up.
    -   All API calls should only be reads. This stops us from having to worry
        about most concurrency issues
    -   Note that printing inside of a thread does not work properly! Use sys.stdout.flush() after the print statement.
        See here: https://stackoverflow.com/questions/18234469/python-multithreaded-print-statements-delayed-until-all-threads-complete-executi
    """

    def __init__(self, steps_manager: StepsManager, mito_backend: MitoWidgetType):
        self.api_queue: Queue = Queue(MAX_QUEUED_API_CALLS)

        # Save some variables for ease
        self.steps_manager = steps_manager
        self.mito_backend = mito_backend

        self.thread: Optional[Thread] = None
        self.had_first_api_call = False

    def start_api_thread(self) -> None:
        # Note that we make the thread a daemon thread, which practically means that when
        # The process that starts this thread terminate, our API will terminate as well.
        self.thread = Thread(
            target=handle_api_event_thread,
            args=(self.api_queue, self.steps_manager, self.mito_backend),
            daemon=True,
        )

        self.thread.start()


    def process_new_api_call(self, event: Dict[str, Any]) -> None:
        """
        We privilege new API calls over old calls, and evict the old ones
        if the API queue is full.

        Because we are using a queue, only events that have not been started
        being processed will get removed.

        If the key 'priority' is in the event, then we handle it in the main
        thread, as we don't want to drop the event. For example, lazy loading
        data has priority!
        """
        # On the first API call , we check if the API should be threaded, and if the it is not already created -- and create it in this case
        global THREADED

        if not self.had_first_api_call:
            if self.thread is None and get_api_should_be_threaded():
                self.start_api_thread()
                THREADED = True
            else:
                THREADED = False
            
            self.had_first_api_call = True

        if THREADED and "priority" not in event:

            if self.api_queue.full():
                # If the queue is full, we drop the first event, and just return a None
                lost_event = self.api_queue.get()

                self.mito_backend.mito_send({"event": "api_response", "id": lost_event["id"], "data": None})

            self.api_queue.put(event)
        else:
            handle_api_event(self.mito_backend.mito_send, event, self.steps_manager)


def handle_api_event_thread(
    queue: Queue, steps_manager: StepsManager, mito_backend: MitoWidgetType
) -> NoReturn:
    """
    This is the worker thread function, that actually is
    responsible for handling at the API call events.

    It lives forever, and just handles events as it
    receives them from the queue
    """
    while True:
        # Note that this blocks when there is nothing in the queue,
        # and waits till there is something there - so no infinite
        # loop as it is waiting!
        event = queue.get()
        # We place the API handling inside of a try catch,
        # because otherwise if an error is thrown, then the entire thread crashes,
        # and then the API never works again
        try:
            handle_api_event(mito_backend.mito_send, event, steps_manager)
        except:
            # Log in error if it occurs
            log_event_processed(event, steps_manager, failed=True)


def handle_api_event(
    send: Callable, event: Dict[str, Any], steps_manager: StepsManager
) -> None:
    """
    Handler for all API calls. Note that any response to the
    API must return the same ID that the incoming message contains,
    so that the frontend knows how to match the responses.
    """
    result: Any = ''
    params = event['params']
    start_time = perf_counter()
    failed = False

    try:
        if event["type"] == "get_path_contents":
            result = get_path_contents(params)
        elif event["type"] == "get_path_join":
            result = get_path_join(params)
        elif event["type"] == "get_dataframe_as_csv":
            result = get_dataframe_as_csv(params, steps_manager)
        elif event["type"] == "get_column_summary_graph":
            result = get_column_summary_graph(params, steps_manager)
        elif event["type"] == "get_column_describe":
            result = get_column_describe(params, steps_manager)
        elif event["type"] == "get_params":
            result = get_params(params, steps_manager)
        elif event["type"] == "get_excel_file_metadata":
            result = get_excel_file_metadata(params, steps_manager)
        elif event["type"] == "get_csv_files_metadata":
            result = get_csv_files_metadata(params, steps_manager)
        elif event["type"] == "get_unique_value_counts":
            result = get_unique_value_counts(params, steps_manager)
        elif event["type"] == "get_split_text_to_columns_preview":
            result = get_split_text_to_columns_preview(params, steps_manager)
        elif event["type"] == "get_dataframe_as_excel":
            result = get_dataframe_as_excel(params, steps_manager)
        elif event["type"] == "get_defined_df_names":
            result = get_defined_df_names(params, steps_manager)
        elif event["type"] == "get_search_matches":
            result = get_search_matches(params, steps_manager)
        elif event["type"] == 'get_imported_files_and_dataframes_from_current_steps':
            result = get_imported_files_and_dataframes_from_current_steps(params, steps_manager)
        elif event["type"] == 'get_imported_files_and_dataframes_from_analysis_name':
            result = get_imported_files_and_dataframes_from_analysis_name(params, steps_manager)
        elif event["type"] == "get_test_imports":
            result = get_test_imports(params, steps_manager)
        elif event["type"] == "get_render_count":
            result = get_render_count(params, steps_manager)
        elif event["type"] == "get_code_snippets":
            result = get_code_snippets(params, steps_manager)
        elif event["type"] == "get_available_snowflake_options_and_defaults":
            result = get_available_snowflake_options_and_defaults(params, steps_manager)
        elif event["type"] == "get_validate_snowflake_credentials":
            result = get_validate_snowflake_credentials(params, steps_manager)
        elif event["type"] == "get_ai_completion":
            result = get_ai_completion(params, steps_manager)
        elif event["type"] == "get_parameterizable_params":
            result = get_parameterizable_params(params, steps_manager)
        elif event["type"] == "get_pr_url_of_new_pr":
            result = get_pr_url_of_new_pr(params, steps_manager)
        # AUTOGENERATED LINE: API.PY CALL (DO NOT DELETE)
        else:
            raise Exception(f"Event: {event} is not a valid API call")

    except:
        failed = True
    
    # Log processing this event (with potential failure)
    log_event_processed(event, steps_manager, failed=failed, start_time=start_time)

    send({"event": "api_response", "id": event["id"], "data": result})
